"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from functools import lru_cache

import jax
from jax.tree_util import tree_flatten, tree_unflatten

from qrisp.jasp.interpreter_tools import extract_invalues, insert_outvalues, eval_jaxpr
from qrisp.jasp.evaluation_tools.buffered_quantum_state import BufferedQuantumState
from qrisp.jasp.primitives import (
    OperationPrimitive,
    AbstractQuantumCircuit,
    AbstractQubitArray,
    AbstractQubit,
)
from qrisp.core import recursive_qv_search
from qrisp.circuit import fast_append


def jaspify(func=None, terminal_sampling=False):
    """
    This simulator is the established Qrisp simulator linked to the Jasp infrastructure.
    Among a variety of simulation tricks, the simulator can leverage state sparsity,
    allowing simulations with up to hundreds of qubits!

    To be called as a decorator of a Jasp-traceable function.

    .. note::

        If you are developing a hybrid algorithm like QAOA or VQE that relies
        heavily on sampling, please activate the ``terminal_sampling`` feature.

    Parameters
    ----------
    func : callable
        The function to simulate.
    terminal_sampling : bool, optional
        Whether to leverage the terminal sampling strategy. Significantly fast
        for all sampling tasks but can yield incorrect results in some situations.
        Check out :ref:`terminal_sampling` form more details. The default is False.

    Returns
    -------
    callable
        A function performing the simulation.

    Examples
    --------

    We simulate a function creating a simple GHZ state:

    ::

        from qrisp import *
        from qrisp.jasp import *

        @jaspify
        def main():

            qf = QuantumFloat(5)

            h(qf[0])

            for i in range(1, 5):
                cx(qf[0], qf[i])

            return measure(qf)

        print(main())
        # Yields either 0 or 31

    To highlight the speed of the terminal sampling feature, we :ref:`sample` from a
    uniform superposition

    ::

        def state_prep():
            qf = QuantumFloat(5)
            h(qf)
            return qf

        @jaspify
        def without_terminal_sampling():
            sampling_func = sample(state_prep, shots = 10000)
            return sampling_func()

        @jaspify(terminal_sampling = True)
        def with_terminal_sampling():
            sampling_func = sample(state_prep, shots = 10000)
            return sampling_func()


    Benchmark the time difference:

    ::

        import time

        t0 = time.time()
        res = without_terminal_sampling()
        print(time.time() - t0)
        # Yields
        # 43.78982925

        t0 = time.time()
        res = with_terminal_sampling()
        print(time.time() - t0)
        # Yields
        # 0.550775527


    """

    if isinstance(func, bool):
        terminal_sampling = func
        func = None

    if func is None:
        return lambda x: jaspify(x, terminal_sampling=terminal_sampling)

    from qrisp.jasp import make_jaspr

    treedef_container = []

    def tracing_function(*args):
        res = func(*args)
        flattened_values, tree_def = tree_flatten(res)
        treedef_container.append(tree_def)
        return flattened_values

    def return_function(*args):
        # To prevent "accidental deletion" induced non-determinism we set the
        # garbage collection mode to manual
        if terminal_sampling:
            garbage_collection = "manual"
        else:
            garbage_collection = "auto"
        jaspr = make_jaspr(tracing_function, garbage_collection=garbage_collection)(
            *args
        )
        jaspr_res = simulate_jaspr(jaspr, *args, terminal_sampling=terminal_sampling)
        if isinstance(jaspr_res, tuple):
            jaspr_res = tree_unflatten(treedef_container[0], jaspr_res)
        if len(recursive_qv_search(jaspr_res)):
            raise Exception("Tried to jaspify function returning a QuantumVariable")
        return jaspr_res

    return return_function


def stimulate(func=None):
    """
    This function leverages the
    `Stim simulator <https://github.com/quantumlib/Stim?tab=readme-ov-file>`_
    to evaluate a Jasp-traceable function containing only Clifford gates.
    Stim is a popular tool to simulate quantum error correction codes.

    .. note::

        To use this simulator, you need stim installed, which can be achieved via
        ``pip install stim``.

    Parameters
    ----------
    func : callable
        The function to simulate.

    Returns
    -------
    callable
        A function performing the simulation.

    Examples
    --------

    We simulate a function creating a simple GHZ state:

    ::

        from qrisp import *
        from qrisp.jasp import *

        @stimulate
        def main():

            qf = QuantumFloat(5)

            h(qf[0])

            for i in range(1, 5):
                cx(qf[0], qf[i])

            return measure(qf)

        print(main())
        # Yields either 0 or 31

    The ``stimulate`` decorator can also simulate real-time features:

    ::

        @stimulate
        def main():

            qf = QuantumFloat(5)

            h(qf[0])

            cl_bl = measure(qf[0])

            with control(cl_bl):
                for i in range(1, 5):
                    x(qf[i])

            return measure(qf)

        print(main())
        # Yields either 0 or 31

    """

    from qrisp.jasp import make_jaspr

    treedef_container = []

    def tracing_function(*args):
        res = func(*args)
        flattened_values, tree_def = tree_flatten(res)
        treedef_container.append(tree_def)
        return flattened_values

    def return_function(*args):
        jaspr = make_jaspr(tracing_function)(*args)
        jaspr_res = simulate_jaspr(jaspr, *args, simulator="stim")
        if isinstance(jaspr_res, tuple):
            jaspr_res = tree_unflatten(treedef_container[0], jaspr_res)
        if len(recursive_qv_search(jaspr_res)):
            raise Exception("Tried to simulate function returning a QuantumVariable")
        return jaspr_res

    return return_function


def simulate_jaspr(
    jaspr, *args, terminal_sampling=False, simulator="qrisp", return_gate_counts=False
):

    from qrisp.alg_primitives.mcx_algs.circuit_library import gidney_qc

    if len(jaspr.outvars) == 1:
        return None

    if simulator == "stim":
        if terminal_sampling:
            raise Exception("Terminal sampling with stim is currently not implemented")
    elif not simulator == "qrisp":
        raise Exception(f"Don't know simulator {simulator}")

    args = list(tree_flatten(args)[0]) + [BufferedQuantumState(simulator)]

    def eqn_evaluator(eqn, context_dic):

        if eqn.primitive.name == "pjit":

            function_name = eqn.params["name"]
            jaxpr = eqn.params["jaxpr"]

            if terminal_sampling:

                translation_dic = {
                    "expectation_value_eval_function": "ev",
                    "sampling_eval_function": "array",
                    "dict_sampling_eval_function": "dict",
                }

                from qrisp.jasp.interpreter_tools import terminal_sampling_evaluator

                if function_name in translation_dic:
                    terminal_sampling_evaluator(translation_dic[function_name])(
                        eqn, context_dic, eqn_evaluator=eqn_evaluator
                    )
                    return

            invalues = extract_invalues(eqn, context_dic)

            # If there are only classical values, we attempt to compile using the jax pipeline
            for var in jaxpr.jaxpr.invars + jaxpr.jaxpr.outvars:
                if isinstance(
                    var.aval,
                    (AbstractQuantumCircuit, AbstractQubitArray, AbstractQubit),
                ):
                    break
            else:

                compiled_function, is_executable = compile_cl_func(
                    jaxpr.jaxpr, function_name
                )

                # Functions with purely classical inputs/outputs can still contain
                # kernelized quantum functions. This will raise an NotImplementedError
                # when attempting to compile. Since the compile_cl_func is lru_cached
                # we can store this information to avoid further attempts at compiling
                # such a function.
                if is_executable[0]:
                    try:
                        outvalues = compiled_function(*(jaxpr.consts + invalues))
                        if len(jaxpr.jaxpr.outvars) > 1:
                            insert_outvalues(eqn, context_dic, outvalues)
                        else:
                            insert_outvalues(eqn, context_dic, [outvalues])
                        return False
                    except NotImplementedError:
                        is_executable[0] = False

            # We simulate the inverse Gidney mcx via the non-hybrid version because
            # the hybrid version prevents the simulator from fusing gates, which
            # slows down the simulation
            if eqn.params["name"] == "gidney_mcx_inv":
                invalues[-1].append(gidney_qc.inverse().to_gate(), invalues[:-1])
                outvalues = [invalues[-1]]
            else:
                outvalues = eval_jaxpr(
                    eqn.params["jaxpr"], eqn_evaluator=eqn_evaluator
                )(*invalues)
            if not isinstance(outvalues, (list, tuple)):
                outvalues = [outvalues]
            insert_outvalues(eqn, context_dic, outvalues)
        elif eqn.primitive.name == "jasp.create_quantum_kernel":
            insert_outvalues(eqn, context_dic, BufferedQuantumState(simulator))
        elif eqn.primitive.name == "jasp.consume_quantum_kernel":
            pass
        else:
            return True

    with fast_append(3):
        res = eval_jaxpr(jaspr, eqn_evaluator=eqn_evaluator)(*(args))

    if return_gate_counts:
        return res[-1].gate_counts

    if len(jaspr.outvars) == 2:
        return res[0]
    else:
        return res[:-1]


@lru_cache(maxsize=int(1e5))
def compile_cl_func(jaxpr, function_name):
    return jax.jit(eval_jaxpr(jaxpr)), [True]



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import types

import numpy as np

from qrisp.circuit import ControlledOperation

"""
TODO: 
-- clbit interaction with measurements 
-- conditional environment, i.e. mid-circuit measurement is possible, have to create this
    --> otherwise no clbit interaction?

BUGS:

"""


def create_qml_instruction(op):
    import pennylane as qml

    # create basic, straight forward instructions
    if op.name == "rxx":
        qml_ins = qml.IsingXX
    elif op.name == "ryy":
        qml_ins = qml.IsingYY
    elif op.name == "rzz":
        qml_ins = qml.IsingZZ

    elif op.name == "p":
        qml_ins = qml.RZ
    elif op.name == "rx":
        qml_ins = qml.RX
    elif op.name == "ry":
        qml_ins = qml.RY
    elif op.name == "rz":
        qml_ins = qml.RZ
    elif op.name == "u1":
        qml_ins = qml.U1

    elif op.name == "u3":
        qml_ins = qml.U3
    elif op.name == "xxyy":
        qml_ins = qml.IsingXY
    elif op.name == "swap":
        qml_ins = qml.SWAP
    elif op.name == "h":
        qml_ins = qml.Hadamard
    elif op.name == "x":
        qml_ins = qml.PauliX
    elif op.name == "y":
        qml_ins = qml.PauliY
    elif op.name == "z":
        qml_ins = qml.PauliZ
    elif op.name == "s" or op.name == "s_dg":
        qml_ins = qml.S
    elif op.name == "t" or op.name == "t_dg":
        qml_ins = qml.T
    elif op.name == "sx" or op.name == "sx_dg":
        qml_ins = qml.SX

    # complex gate, with subcircuit definition we create a representative subcircuit
    elif op.definition:
        circ = qml_converter(op.definition, circ=True)
        return circ
    else:
        Exception("Cannot convert")
    return qml_ins


import sympy


def qml_converter(qc, circ=False):
    import pennylane as qml

    def circuit(*args, wires=None):

        circFlag = False

        qbit_dic = {}
        # add qubits
        if wires is None:
            for qubit in qc.qubits:
                qbit_dic[qubit.identifier] = qubit.identifier
        else:
            for i in range(len(qc.qubits)):
                qbit_dic[qc.qubits[i].identifier] = wires[i]

        # save measurements
        measurelist = []
        symbols = list(qc.abstract_params)

        for index in range(len(qc.data)):

            op = qc.data[index].op
            params = list(op.params)
            for i in range(len(params)):
                if isinstance(params[i], sympy.Symbol):
                    params[i] = args[symbols.index(params[i])]

            qubit_list = [qbit_dic[qubit.identifier] for qubit in qc.data[index].qubits]
            op_qubits = qc.data[index].qubits

            if op.name in ["qb_alloc", "qb_dealloc"]:
                continue

            if op.name == "measure":
                ### below: possible implementation for nested circuit measurments
                # if circ == True:
                #   Exception("Measurements in nested circuits not supported (yet)")
                measurelist.append(*qubit_list)
                continue

            ###### BugCatcher + special gates ######
            if op.name in ["sx", "sx_dg", "id", "t", "t_dg", "s", "s_dg"]:
                # bugged -> params empty
                params = []
            if op.name == "xxyy":
                # deffed as RXXYY with angle pi
                params = [np.pi]

            if op.name == "cx":
                # maybe adjustment necessary here
                if hasattr(op, "ctrl_state"):
                    qml_ins = qml.CNOT([qubit_list[0], qubit_list[1]])
                else:
                    qml_ins = qml.CNOT([qubit_list[0], qubit_list[1]])

            elif op.name == "cy":
                if hasattr(op, "ctrl_state"):
                    qml_ins = qml.CY([qubit_list[0], qubit_list[1]])
                else:
                    qml_ins = qml.CY([qubit_list[0], qubit_list[1]])

            elif op.name == "cz":
                if hasattr(op, "ctrl_state"):
                    qml_ins = qml.CZ([qubit_list[0], qubit_list[1]])
                else:
                    qml_ins = qml.CZ([qubit_list[0], qubit_list[1]])

            elif op.name == "cp":
                if hasattr(op, "ctrl_state"):
                    qml_ins = qml.ControlledPhaseShift(
                        *params, [qubit_list[0], qubit_list[1]]
                    )
                else:
                    qml_ins = qml.ControlledPhaseShift(
                        *params, [qubit_list[0], qubit_list[1]]
                    )

            # return the controlled instruction or the nested circuit
            elif isinstance(op, ControlledOperation):
                qml_trafo = create_qml_instruction(op)
                if isinstance(qml_trafo, types.FunctionType):
                    circFlag = True
                    wire_map = dict()
                    for i in range(len(op_qubits)):
                        # wire_map[i] = op_qubits[i].identifier
                        wire_map[op.definition.qubits[i].identifier] = op_qubits[
                            i
                        ].identifier
                    mapped_quantum_function = qml.map_wires(qml_trafo, wire_map)
                    qml_ins = qml.ctrl(
                        mapped_quantum_function, control=op.ctrl_state[::-1]
                    )
                else:
                    qml_ins = qml.ctrl(qml_trafo, control=op.ctrl_state[::-1])

            # return the instruction or the nested circuit
            else:
                qml_trafo = create_qml_instruction(op)
                if isinstance(qml_trafo, types.FunctionType):
                    circFlag = True
                    wire_map = dict()
                    for i in range(len(op_qubits)):
                        # map the wires as given for the nested circuit
                        wire_map[op.definition.qubits[i].identifier] = op_qubits[
                            i
                        ].identifier
                    mapped_quantum_function = qml.map_wires(qml_trafo, wire_map)

                else:
                    if "_dg" in op.name:
                        qml_ins = qml.adjoint(qml_trafo(*params, wires=qubit_list))
                    else:
                        qml_ins = qml_trafo(*params, wires=qubit_list)

            if not isinstance(circ, bool):
                circ()
            elif not circFlag:
                qml_ins
            else:
                mapped_quantum_function()

        # add measurements at the end, i.e. return a count object
        if len(measurelist):
            return qml.counts(wires=measurelist)

    return circuit



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.alg_primitives.mcx_algs.multi_cx import *
from qrisp.alg_primitives.mcx_algs.amy import *
from qrisp.alg_primitives.mcx_algs.balauca import *
from qrisp.alg_primitives.mcx_algs.gidney import *
from qrisp.alg_primitives.mcx_algs.gms import *
from qrisp.alg_primitives.mcx_algs.gray import *
from qrisp.alg_primitives.mcx_algs.gray_pt import *
from qrisp.alg_primitives.mcx_algs.jones import *
from qrisp.alg_primitives.mcx_algs.maslov import *
from qrisp.alg_primitives.mcx_algs.yong import *
from qrisp.alg_primitives.mcx_algs.khattar import *


# Interface function to quickly change between different implementations of
# multi controlled not gates
def multi_cx(n, method=None):
    # from qrisp.circuit import transpile

    if method == "gms":
        return gms_multi_cx(n)

    elif method in ["gray_pt"]:
        return pt_multi_cx(n)

    elif method == "gray_pt_inv":
        return pt_multi_cx(n).inverse()

    elif method in ["gray", "auto", None]:
        return gray_multi_cx(n)

    elif method == "gidney":
        return GidneyLogicalAND()
    elif method == "gidney_inv":
        return GidneyLogicalAND(inv=True)
    else:
        raise Exception('method "' + method + '" not implemented')



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import inspect

from jax.lax import while_loop, cond
import jax
import jax.numpy as jnp

from qrisp.circuit import XGate
from qrisp.jasp import (
    TracingQuantumSession,
    AbstractQubitArray,
    DynamicQubitArray,
    qache,
)
from qrisp.jasp.primitives import (
    Measurement_p,
    OperationPrimitive,
    get_qubit_p,
    get_size_p,
    delete_qubits_p,
    reset_p,
)


def RUS(*trial_function, **jit_kwargs):
    r"""
    Decorator to deploy repeat-until-success (RUS) components. At the core,
    RUS repeats a given quantum subroutine followed by a qubit measurement until
    the measurement returns the value ``1``. This step is prevalent
    in many important algorithms, among them the
    `HHL algorithm <https://arxiv.org/abs/0811.3171>`_ or the
    `LCU procedure <https://arxiv.org/abs/1202.5822>`_.

    Within Jasp, RUS steps can be realized by providing the quantum subroutine
    as a "trial function", which returns a boolean value (the repetition condition) and
    possibly other return values.

    It is important to note that the trial function can not receive quantum
    arguments. This is because after each trial, a new copy of these arguments
    would be required to perform the next iteration, which is prohibited by
    the no-clone theorem. It is however legal to provide classical arguments.

    Parameters
    ----------
    trial_function : callable
        A function returning a boolean value as the first return value. More
        return values are possible.
    static_argnums : int or list[int], optional
        A list of integers specifying which arguments are considered static in
        the sense of `jax.jit <https://jax.readthedocs.io/en/latest/_autosummary/jax.jit.html>`_.
        The first argument is indicated by 1, the second by 2, etc. The default
        is ``[]``.
    static_argnames : str or list[str], optional
        A list of strings specifying which arguments are considered static in
        the sense of `jax.jit <https://jax.readthedocs.io/en/latest/_autosummary/jax.jit.html>`_.
        The default is ``[]``.

    Returns
    -------
    callable
        A function that performs the RUS protocol with the trial function. The
        return values of this function are the return values of the trial function
        WITHOUT the boolean value.

    Examples
    --------

    To demonstrate the RUS behavior, we initialize a GHZ state

    .. math::

        \ket{\psi} = \frac{\ket{00000} + \ket{11111}}{\sqrt{2}}

    and measure the first qubit into a boolean value. This will be the value
    to cancel the repetition. This will collapse the GHZ state into either
    $\ket{00000}$ (which will cause a new repetition) or $\ket{11111}$, which
    cancels the loop. After the repetition is canceled we are therefore
    guaranteed to have the latter state.

    ::

        from qrisp.jasp import RUS, make_jaspr
        from qrisp import QuantumFloat, h, cx, measure

        @RUS
        def rus_trial_function():
            qf = QuantumFloat(5)
            h(qf[0])

            for i in range(1, 5):
                cx(qf[0], qf[i])

            cancelation_bool = measure(qf[0])
            return cancelation_bool, qf

        def call_RUS_example():

            qf = rus_trial_function()

            return measure(qf)

    Create the ``jaspr`` and simulate:

    ::

        jaspr = make_jaspr(call_RUS_example)()
        print(jaspr())
        # Yields, 31 which is the decimal version of 11111

    **Static arguments**

    To demonstrate the specification of static arguments, we will realize implement a
    simple `linear combination of unitaries <https://arxiv.org/abs/1202.5822>`_.

    Our implementation initializes a state of the form

    .. math::

        \left( \sum_{i = 0}^N c_i U_i \right) \ket{0}.

    We achieve this by specifying a set of unitaries $U_i$ in the form of a
    tuple of functions, each processing a :ref:`QuantumFloat`.

    The coefficients $c_i$ are specified through a function preparing the state

    .. math::

        \ket{\psi} = \sum_{i = 0}^N c_i \ket{i}

    For the state preparation function we specify two options to experiment with.
    A two qubit uniform superposition and a function that brings only the first
    qubit into superpostion.

    ::

        def state_prep_full(qv):
            h(qv[0])
            h(qv[1])

        def state_prep_half(qv):
            h(qv[0])

    For the first one we have $c_0 = c_1 = c_2 = c_3 = \sqrt{0.25}$. The second one
    gives $c_0 = c_1 = \sqrt{0.5}$ and $c_2 = c_3 = 0$.

    The next step is to define the unitaries $U_i$ in the form of a tuple
    of functions.

    ::

        from qrisp.jasp import *
        from qrisp import *

        def case_function_0(x):
            x += 3

        def case_function_1(x):
            x += 4

        def case_function_2(x):
            x += 5

        def case_function_3(x):
            x += 6

        case_functions = (case_function_0,
                          case_function_1,
                          case_function_2,
                          case_function_3)


    These functions each represent the unitary:

    .. math::

        U_i \ket{0} = \ket{i+3}

    Executing a linear combination of unitaries therefore gives

    .. math::

        \left( \sum_{i = 0}^N c_i U_i \right) \ket{0} = \sum_{i = 0}^N c_i \ket{i+3}


    Now we implement the LCU procedure.

    ::

        # Specify the corresponding arguments of the block encoding as "static",
        # i.e. compile time constants.

        @RUS(static_argnums = [2,3])
        def block_encoding(return_size, state_preparation, case_functions):

            # This QuantumFloat will be returned
            qf = QuantumFloat(return_size)

            # Specify the QuantumVariable that indicates, which
            # case to execute
            n = int(np.ceil(np.log2(len(case_functions))))
            case_indicator = QuantumFloat(n)

            # Turn into a list of qubits
            case_indicator_qubits = [case_indicator[i] for i in range(n)]

            # Perform the LCU protocoll
            with conjugate(state_preparation)(case_indicator):
                for i in range(len(case_functions)):
                    with control(case_indicator_qubits, ctrl_state = i):
                        case_functions[i](qf)

            # Compute the success condition
            success_bool = (measure(case_indicator) == 0)

            return success_bool, qf

    Finally, evaluate via the :ref:`terminal_sampling <terminal_sampling>`
    feature:

    ::

        @terminal_sampling
        def main():
            return block_encoding(4, state_prep_full, case_functions)

        print(main())
        # Yields: {3.0: 0.25, 4.0: 0.25, 5.0: 0.25, 6.0: 0.25}

    Evaluate the other state preparation function

    ::

        @terminal_sampling
        def main():
            return block_encoding(4, state_prep_half, case_functions)

        print(main())
        # Yields: {3.0: 0.5, 4.0: 0.5}

    As expected, the full state preparation function yields a state proportional
    to

    .. math::

        \ket{3} + \ket{4} + \ket{5} + \ket{6}.

    The second state preparation gives us

    .. math::

        \ket{3} + \ket{4}.

    """
    if len(trial_function) == 0:
        return lambda x: RUS(x, **jit_kwargs)
    else:
        trial_function = trial_function[0]

    # The idea for implementing this feature is to execute the function once qached
    # to collect the output QuantumVariable object.
    # From the infered output signature the q_while_loop is constructed

    def return_function(*trial_args):
        
        from qrisp.jasp import q_while_loop, q_cond
        from qrisp.core import recursive_qv_search, reset


        abs_qs = TracingQuantumSession.get_instance()

        initial_gc_mode = abs_qs.gc_mode
        
        # Set the garbage collection mode to temporarily auto to collect 
        # any ancillas that have not been deleted.
        abs_qs.gc_mode = "auto"
        # Execute the function
        qached_function = qache(trial_function, **jit_kwargs)
        first_iter_res = qached_function(*trial_args)

        abs_qs.gc_mode = initial_gc_mode
        
        
        # Filter out the static arguments
        if "static_argnums" in jit_kwargs:
            static_argnums = jit_kwargs["static_argnums"]
            if isinstance(static_argnums, int):
                static_argnums = [static_argnums]
        else:
            static_argnums = []
        
        if "static_argnames" in jit_kwargs:
            argname_list = inspect.getfullargspec(trial_function)
            for i in range(len(argname_list)):
                if argname_list[i] in jit_kwargs["static_argnames"]:
                    static_argnums.append(i)
        
        
        dynamic_args = []
        
        for i in range(len(trial_args)):
            if i not in static_argnums:
                dynamic_args.append(trial_args[i])

        n_arg_vals = len(dynamic_args)        
        
        # Next we construct the body of the loop
        # The q_while_loop receives a tuple of arguments and 
        # also returns a tuple with the same signature.
        # We therefore combine the results of the first iteration with
        # the arguments to execute the loop.

        combined_args = tuple(list(dynamic_args) + list(first_iter_res))

        # This is the body function of the while loop
        def body_fun(args):
            
            # The first step is to reset and delete the results
            # from the previous iteration
            qv_results = recursive_qv_search(args[n_arg_vals:])
            
            abs_qs = TracingQuantumSession.get_instance()
            for qv in qv_results:
                abs_qs.register_qv(qv, None)
                reset(qv)
                qv.delete()

            
            # We now construct the arguments for the function call of
            # the current iteration.
            # For this, we combine the dynamic arguments with the static
            # arguments.
            dynamic_args = list(args[:n_arg_vals])
            
            new_trial_args = []
            for i in range(len(trial_args)):
                if i not in static_argnums:
                    new_trial_args.append(dynamic_args.pop(0))
                else:
                    new_trial_args.append(trial_args[i])

            # Set the garbage collection mode to auto and call the function            
            abs_qs.gc_mode = "auto"
            trial_res = qached_function(*new_trial_args)
            abs_qs.gc_mode = initial_gc_mode
            
            # Update the tuple with initial args and the new results
            combined_args = tuple(list(args[:n_arg_vals]) + list(trial_res))
            return combined_args

        # This is the loop cancelation condition
        def cond_fun(val):
            # The loop cancelation index is located at the second position of the
            # return value tuple
            return ~val[n_arg_vals]

        # We now evaluate the loop

        # If the first iteration was already successful, we simply return the results
        # To realize this behavior we use a q_cond primitive
        def true_fun(combined_args):
            return combined_args

        # If the first iteration was not successfull, we start the loop
        def false_fun(combined_args):
            # Here is the while_loop
            return q_while_loop(cond_fun, body_fun, init_val=combined_args)

        # Evaluate everything
        combined_res = q_cond(first_iter_res[0], true_fun, false_fun, combined_args)
        
        # Return the results
        if len(first_iter_res) == 2:
            return combined_res[n_arg_vals+1]
        else:
            return combined_res[n_arg_vals+1:]

    return return_function



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""


def amplitude_amplification(
    args,
    state_function,
    oracle_function,
    kwargs_oracle={},
    iter=1,
    reflection_indices=None,
):
    r"""
    This method performs `quantum amplitude amplification <https://arxiv.org/abs/quant-ph/0005055>`_.

    The problem of quantum amplitude amplification is described as follows:

    * Given a unitary operator :math:`\mathcal{A}`, let :math:`\ket{\Psi}=\mathcal{A}\ket{0}`.
    * Write :math:`\ket{\Psi}=\ket{\Psi_1}+\ket{\Psi_0}` as a superposition of the orthogonal good and bad components of :math:`\ket{\Psi}`.
    * Enhance the probability :math:`a=\langle\Psi_1|\Psi_1\rangle` that a measurement of $\ket{\Psi}$ yields a good state.

    Let $\theta_a\in [0,\pi/2]$ such that $\sin^2(\theta_a)=a$. Then the amplitude amplification operator $\mathcal Q$ acts as

    .. math::

        \mathcal Q^j\ket{\Psi}=\frac{1}{\sqrt{a}}\sin((2j+1)\theta_a)\ket{\Psi_1}+\frac{1}{\sqrt{1-a}}\cos((2j+1)\theta_a)\ket{\Psi_0}.

    Therefore, after $m$ iterations the probability of measuring a good state is $\sin^2((2m+1)\theta_a)$.

    Parameters
    ----------

    args : QuantumVariable or list[QuantumVariable]
        The (list of) QuantumVariables which represent the state,
        the amplitude amplification is performed on.
    state_function : function
        A Python function preparing the state $\ket{\Psi}$.
        This function will receive the variables in the list ``args`` as arguments in the
        course of this algorithm.
    oracle_function : function
        A Python function tagging the good state $\ket{\Psi_1}$.
        This function will receive the variables in the list ``args`` as arguments in the
        course of this algorithm.
    kwargs_oracle : dict, optional
        A dictionary containing keyword arguments for the oracle. The default is {}.
    iter : int, optional
        The amount of amplitude amplification iterations to perform. The default is 1.
    refection_indices : list[int], optional
        A list indicating with respect to which variables the reflection within the diffuser is performed, i.e. oblivious amplitude amplification is performed.
        By default, the reflection is performed with respect to all variables in ``args``, i.e. standard amplitude amplification is performed.

    Examples
    --------

    We define a function that prepares the state :math:`\ket{\Psi}=\cos(\frac{\pi}{16})\ket{0}+\sin(\frac{\pi}{16})\ket{1}`
    and an oracle that tags the good state :math:`\ket{1}`. In this case, we have :math:`a=\sin^2(\frac{\pi}{16})\approx 0.19509`.

    ::

        from qrisp import z, ry, QuantumBool, amplitude_amplification
        import numpy as np

        def state_function(qb):
            ry(np.pi/8,qb)

        def oracle_function(qb):
            z(qb)

        qb = QuantumBool()

        state_function(qb)

    >>> qb.qs.statevector(decimals=5)
    0.98079∣False⟩+0.19509∣True⟩

    We can enhance the probability of measuring the good state with amplitude amplification:

    >>> amplitude_amplification([qb], state_function, oracle_function)
    >>> qb.qs.statevector(decimals=5)
    0.83147*|False> + 0.55557*|True>

    >>> amplitude_amplification([qb], state_function, oracle_function)
    >>> qb.qs.statevector(decimals=5)
    0.55557*|False> + 0.83147*|True>

    >>> amplitude_amplification([qb], state_function, oracle_function)
    >>> qb.qs.statevector(decimals=5)
    0.19509*|False> + 0.98079*|True>

    """

    from qrisp.grover import diffuser
    from qrisp import merge, recursive_qs_search, IterationEnvironment
    from qrisp.jasp import check_for_tracing_mode, jrange

    if check_for_tracing_mode():
        for i in jrange(iter):
            oracle_function(*args, **kwargs_oracle)
            diffuser(
                args,
                state_function=state_function,
                reflection_indices=reflection_indices,
            )
    else:
        merge(args)
        qs = recursive_qs_search(args)[0]
        if iter > 0:
            with IterationEnvironment(qs, iter):
                oracle_function(*args, **kwargs_oracle)
                diffuser(
                    args,
                    state_function=state_function,
                    reflection_indices=reflection_indices,
                )



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp.qtypes.quantum_float import QuantumFloat
from qrisp.alg_primitives import QFT
from qrisp.alg_primitives.arithmetic import multi_controlled_U_g, hybrid_mult
from qrisp.core.gate_application_functions import h, cx, swap, mcx
from qrisp.environments import conjugate, control
from qrisp.alg_primitives.arithmetic.modular_arithmetic.mod_tools import (
    modinv,
    montgomery_encoder,
)


def QREDC(t, N, m):

    u = QuantumFloat(0, qs=t.qs, name="u*")

    for i in range(m):

        h(t[0])

        transfer_lsb(t, u)

        t.exp_shift(-1)
        multi_controlled_U_g(t, [u[-1]], -N / 2)

    QFT(t, inv=True, exec_swap=False)

    with conjugate(QFT)(t[:-1], exec_swap=False):

        multi_controlled_U_g(t[:-1], [t.sign()], N)

    cx(t[0], t.sign())

    sgn = t.reg.pop(-1)
    t.signed = False

    u.reg.insert(len(u), sgn)

    t.m = t.m - m

    return t, u


def find_optimal_m(b, N):
    m = 0
    n = int(np.ceil(np.log2(N)))

    while True:
        b_trial = montgomery_encoder(b, 2**m, N) % N
        if (N - 1) * b_trial < N * 2 ** (m):
            return m
        m += 1


def montgomery_mod_semi_mul(a, b, output_qg=None, permeable_if_zero=False):

    N = a.modulus

    b = b % N

    m = find_optimal_m(b, N)

    if b == 0:
        if output_qg is None:
            return a.duplicate()
        else:
            return output_qg
    if b == 1:
        if output_qg is None:
            return a.duplicate(init=True)
        else:
            output_qg += a
            return output_qg

    m = find_optimal_m(b, N)
    b = montgomery_encoder(b, 2**m, N) % N

    if output_qg is None:
        t = QuantumFloat(a.size + m, signed=True)

        h(t)
    else:
        if output_qg.modulus != N:
            raise Exception("Output QuantumModulus has incompatible modulus")

        output_qg.extend(m, 0)
        output_qg.add_sign()

        QFT(output_qg, exec_swap=False)

        t = output_qg

    for i in range(a.size):
        multi_controlled_U_g(t, [a[i]], (2**i * b))

    from qrisp import QuantumModulus

    t.__class__ = QuantumModulus
    t.modulus = a.modulus
    t.m = a.m + m
    t.inpl_adder = a.inpl_adder

    return montgomery_red(t, a, b, N, m, permeable_if_zero=permeable_if_zero)


def montgomery_red(t, a, b, N, m, permeable_if_zero=False):

    t, u = QREDC(t, N, m)

    with conjugate(QFT)(u, exec_swap=False):
        if isinstance(b, QuantumFloat):

            hybrid_mult(
                a,
                b,
                output_qf=u,
                init_op=None,
                terminal_op=None,
                cl_factor=-modinv(N, 2 ** (m + 1)),
            )

        else:
            for i in range(len(a)):
                multi_controlled_U_g(u, [a[i]], -((2**i * b)) * modinv(N, 2 ** (m + 1)))

    if permeable_if_zero:
        mcx(list(a) + [t[0]], u[-1], ctrl_state="0" * len(a) + "1", method="balauca")

    u.delete(verify=False)

    return t


from qrisp import merge


def montgomery_mod_mul(a, b, output_qg=None):

    m = int(np.ceil(np.log2((a.modulus - 1) ** 2) + 1)) - a.size

    if a.modulus != b.modulus:
        raise Exception("Tried to multiply two QuantumModulus with differing modulus")

    if output_qg is None:
        t = QuantumFloat(a.size + m, signed=True)
        h(t)

    else:
        if output_qg.modulus != a.modulus:
            raise Exception("Output QuantumModulus has incompatible modulus")

        merge(output_qg.qs, a.qs)
        output_qg.extend(m, 0)
        output_qg.add_sign()
        output_qg.reg.insert(0, output_qg.reg.pop(-1))

        QFT(output_qg, exec_swap=False)

        t = output_qg

    t = hybrid_mult(a, b, init_op=None, terminal_op=None, output_qf=t)

    from qrisp import QuantumModulus

    t.__class__ = QuantumModulus
    t.modulus = a.modulus
    t.m = a.m + b.m
    t.inpl_adder = a.inpl_adder

    t = montgomery_red(t, a, b, a.modulus, m)

    return t


from qrisp import invert
from qrisp.environments import custom_control


@custom_control
def qft_semi_cl_inpl_mult(a, X, ctrl=None, treat_invalid=False):

    X = X % a.modulus

    if X == 0:
        raise Exception(
            "Tried to perform in-place multiplication with 0 (not invertible)"
        )
    if X == 1:
        return a

    tmp = a.duplicate(qs=a.qs)

    from qrisp import multi_measurement, less_than, redirect_qfunction, QuantumModulus

    if treat_invalid:
        a.__class__ = QuantumFloat
        reduced = a < a.modulus
        a.__class__ = QuantumModulus
        ctrl = [ctrl, reduced[0]]

    if ctrl is not None:
        with control(ctrl, invert=True):
            swap(tmp, a)

    tmp = montgomery_mod_semi_mul(
        a, X, output_qg=tmp, permeable_if_zero=ctrl is not None
    )

    if ctrl is not None:
        with control(ctrl, invert=True):
            swap(tmp, a)

    with invert():
        a = montgomery_mod_semi_mul(
            tmp,
            modinv(X, a.modulus) % a.modulus,
            output_qg=a,
            permeable_if_zero=ctrl is not None,
        )

    if ctrl is not None:
        with control(ctrl, invert=False):
            swap(tmp, a)
    else:
        swap(tmp, a)

    tmp.delete(verify=False)

    if treat_invalid:
        a.__class__ = QuantumFloat
        redirect_qfunction(less_than)(a, a.modulus, target=reduced)
        a.__class__ = QuantumModulus
        reduced.delete(verify=False)

    return a


def transfer_lsb(from_qv, to_qv):
    lsb = from_qv.reg.pop(0)
    from_qv.exponent += 1

    to_qv.reg.insert(len(to_qv), lsb)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import (
    QuantumVariable,
    h,
    barrier,
    rz,
    ry,
    rx,
    cx,
    QuantumArray,
    xxyy,
    p,
    invert,
    conjugate,
    mcp,
    auto_uncompute,
    control,
)
from qrisp.jasp import jrange


def RX_mixer(qv, beta):
    """
    Applies an RX gate to each qubit in ``qv``.

    The RX gate is a single-qubit rotation about the x-axis. It is used as a mixer in QAOA to drive transitions between different states.

    Parameters
    ----------
    qv : :ref:`QuantumVariable`
        The quantum variable to which the RX gate is applied.
    beta : float or sympy.Symbol
        The phase shift value for the RX gate.

    """
    rx(2 * beta, qv)


def RY_mixer(qv, beta):
    """
    Applies an RY gate to each qubit in ``qv``.

    The RY gate is a single-qubit rotation about the x-axis. It is used as a mixer in QAOA to drive transitions between different states.

    Parameters
    ----------
    qv : :ref:`QuantumVariable`
        The quantum variable to which the RY gate is applied.
    beta : float or sympy.Symbol
        The phase shift value for the RY gate.

    """
    ry(2 * beta, qv)


def XY_mixer(qv, beta):
    """
    Applies multiple XX+YY gates to ``qv`` such that each qubit has interacted with its neighbour at least once.

    The XX+YY gate is a two-qubit gate that performs rotations around the XY plane. It is used as a mixer in QAOA to drive transitions between different states.

    A defining feature of this mixer is the fact, that it keeps the number of ones (or equivalently zeros) in the binary representation of the state invariant.

    Parameters
    ----------
    qv : :ref:`QuantumVariable`
        The quantum variable to which the XY gate is applied.
    beta : float or sympy.Symbol
        The phase shift value for the XY gate.

    """
    N = qv.size

    for i in jrange(0, N // 2):
        q1 = qv[2 * i]
        q2 = qv[2 * i + 1]
        xxyy(4 * beta, 0, q1, q2)

    for i in jrange(0, (N - 2 + N % 2) // 2):
        q1 = qv[2 * i + 1]
        q2 = qv[2 * i + 2]
        xxyy(4 * beta, 0, q1, q2)

    xxyy(4 * beta, 0, qv[N - 1], qv[0])


def apply_XY_mixer(quantumcolor_array, beta):
    for qcolor in quantumcolor_array:
        XY_mixer(qcolor, beta)
    return quantumcolor_array


def RZ_mixer(qv, beta):
    """
    This function applies an RZ gate with a negative phase shift to a given quantum variable.

    Parameters
    ----------
    qv : :ref:`QuantumVariable`
        The quantum variable to which the RZ gate is applied.
    beta : float or sympy.Symbol
        The phase shift value for the RZ gate.

    """
    rz(-beta, qv)


def grover_mixer(qv, beta):
    """
    Performs the parametrized Grover diffuser.

    Parameters
    ----------
    qv : :ref:`QuantumVariable`
        The QuantumVariable to be mixed.
    beta : float or sympy.Symbol
        The mixing parameter.

    """
    from qrisp.grover import diffuser

    diffuser(qv, phase=beta)


def constrained_mixer_gen(constraint_oracle, winner_state_amount):
    r"""
    Generates a customized mixer function that leaves arbitrary constraints intact.
    The constraints are specified via a ``constraint_oracle`` function, which
    is taking a :ref:`QuantumVariable` or :ref:`QuantumArray` and apply a phase $\phi$
    (specified by the keyword argument ``phase``) to the states that are allowed
    by the constraints.

    Additionally the amount of winner states needs to be known. For this the user
    needs to provide the function ``winner_state_amount``, that returns the number
    of winner states for a given qubit amount. This number can be an approximation,
    however faulty values can cause leakage into the state-space that is forbidden
    by the constraints.

    For more details regarding implementation specifics please check the
    corresponding :ref:`tutorial <QAOA101>`.

    Parameters
    ----------
    constraint_oracle : function
        A function of a :ref:`QuantumVariable` or :ref:`QuantumArray`. Also needs to
        support the keyword argument ``phase``. This function should apply the phase
        specified by the keyword argument to the allowed states.
    winner_state_amount : function
        A function of a QuantumVariable or QuantumArray, that returns the amount
        of winner states for that QuantumVariable.


    Returns
    -------
    constrained_mixer : function
        A mixer function that does not leave the allowed space specified by the oracle.

    Examples
    --------

    We create a mixer function that only mixes among the states where the first and the
    last qubit disagree. In more mathematical terms - they satisfy the following
    constraint function.


    .. math::

        f: \mathbb{F}_2^n \rightarrow \mathbb{F}_2, x \rightarrow (x_{n-1} \neq x_0)

    ::

        from qrisp.qaoa import constrained_mixer_gen
        from qrisp import QuantumVariable, auto_uncompute, cx, p

        @auto_uncompute
        def constraint_oracle(qarg, phase):

            predicate = QuantumBool()

            cx(qarg[0], predicate)
            cx(qarg[-1], predicate)
            p(phase, predicate)

        def winner_state_amount(qarg):
            return 2**(len(qarg) - 1)

        mixer = constrained_mixer_gen(constraint_oracle, winner_state_amount)

    To test the mixer, we create a :ref:`QuantumVariable`:

    ::

        import numpy as np
        beta = np.pi

        qv = QuantumVariable(3)
        qv[:] = "101"
        mixer(qv, beta)
        print(qv)
        #Yields: {'101': 1.0}
        #Leaves forbidden states invariant

        qv = QuantumVariable(3)
        qv[:] = "100"
        mixer(qv, beta)
        print(qv)
        #Yields: {'100': 0.25, '110': 0.25, '001': 0.25, '011': 0.25}
        #Only mixes among allowed states


    """

    from qrisp.grover import grovers_alg

    def prep_psi(qarg):

        if isinstance(qarg, QuantumVariable):
            qubit_amount = len(qarg)
        elif isinstance(qarg, QuantumArray):
            qubit_amount = len(qarg.qtype) * len(qarg.flatten())
        else:
            raise Exception(
                f"Argument type {type(qarg)} not supported for constrained mixer"
            )

        grovers_alg(
            qarg,
            constraint_oracle,
            exact=True,
            winner_state_amount=winner_state_amount(qarg),
        )

    def inv_prep_psi(qarg):

        with invert():
            prep_psi(qarg)

    def constrained_mixer(qarg, beta):

        with conjugate(inv_prep_psi)(qarg):
            mcp(beta, qarg, ctrl_state=0)

    return constrained_mixer


def controlled_RX_mixer_gen(predicate):
    r"""
    Generate a controlled RX mixer for a given predicate function.

    Parameters
    ----------
    predicate : function
        A function receiving a ``QuantumVariable`` and an index $i$.
        This function returns a ``QuantumBool`` indicating if the predicate is satisfied for ``qv[i]``,
        that is, if the element ``qv[i]`` should be swapped in.

    Returns
    -------
    controlled_RX_mixer : function
        A function receiving a ``QuantumVariable`` and a real parameter $\beta$.
        This function performs the application of the mixing operator.

    Examples
    --------

    We define the predicate function for the :ref:`MaxIndepSet <maxIndepSetQAOA>` problem. It returns ``True`` for the index (node) $i$ if
    all neighbors $j$ of the node $i$ in the graph $G$ are not selected, and ``False`` otherwise.

    ::

        from qrisp import QuantumVariable, QuantumBool, h, mcx, auto_uncompute, multi_measurement
        import networkx as nx

        G = nx.Graph()
        G.add_edges_from([(0, 1), (1, 2), (2, 0)])
        neighbors_dict = {node: list(G.adj[node]) for node in G.nodes()}

        def predicate(qv,i):
            qbl = QuantumBool()
            if len(neighbors_dict[i])==0:
                x(qbl)
            else:
                mcx([qv[j] for j in neighbors_dict[i]],qbl,ctrl_state='0'*len(neighbors_dict[i]))
            return qbl

        qv = QuantumVariable(3)
        h(qv)
        qbl = predicate(qv,0)
        multi_measurement([qv,qbl])
        # Yields: {('000', True): 0.125,('100', True): 0.125,('010', False): 0.125,('110', False): 0.125,('001', False): 0.125,('101', False): 0.125,('011', False): 0.125,('111', False): 0.125}

    The resulting ``controlled_RX_mixer`` then only swaps the node $i$ in if all neighbors $j$ in the graph $G$ are not selected.

    """

    @auto_uncompute
    def controlled_RX_mixer(qv, beta):
        m = qv.size
        for i in range(m):
            with control(predicate(qv, i)):
                rx(beta, qv[i])

    return controlled_RX_mixer


""" from qrisp import as_hamiltonian
@as_hamiltonian
def mcp_as_hamiltonian(qv, beta):
    if qv == "0001":
        p(beta, qv)
    elif qv == "0011":
        p(beta, qv)
    elif qv == "0111":
        p(beta, qv)
    elif qv == "1111":
        p(beta, qv) """


# formulate on q_array
def portfolio_mixer():
    """
    Multi-Channel constrained mixer to be applied for a discrete portfolio rebalancing problem, as seen in https://arxiv.org/pdf/2006.00354.pdf.
    This Mixer keeps the constraints in terms of lots on the portfolio intact. This is achieved by mixing between Dicke States.


    Returns:
    --------

    apply_mixer : function
        The Mixer to be applied to a QuantumVariable

    Examples:
    ---------

    We initiate a QuantumVariable in the "0011" state and from this partially mix into the Dicke state space with Hamming weight 2.

    ::

        from qrisp import QuantumVariable, x
        import numpy as np

        qv = QuantumVariable(4)
        x(qv[2])
        x(qv[3])

        from qrisp.qaoa.mixers import portfolio_mixer
        mixer_op = portfolio_mixer()
        mixer_op(qv, np.pi/8)

    """
    from qrisp.alg_primitives import dicke_state

    def inv_prepare_dicke(qv, k):
        with invert():
            dicke_state(qv, k)

    def apply_mixer(q_array, beta):
        half = int(len(q_array[0]))
        qv1 = q_array[0]
        qv2 = q_array[1]

        # omfg this is harcoded-- problematic one

        with conjugate(inv_prepare_dicke)(qv1, half):
            # mehrere mcp-gates, as hamiltonian
            # mcp_as_hamiltonian(qv1, beta=beta)
            for i in range(half):
                ctrl_state = "0" * (half - i - 1) + ("1" * (i + 1))
                # print(ctrl_state)
                mcp(beta, qv1, ctrl_state=ctrl_state)
            """ mcp(beta, qv1, ctrl_state = "0001")
            mcp(beta, qv1, ctrl_state = "0011")
            mcp(beta, qv1, ctrl_state = "0111")
            mcp(beta, qv1, ctrl_state = "1111") """

        with conjugate(inv_prepare_dicke)(qv2, half):
            for i in range(half):
                ctrl_state = "0" * (half - i - 1) + ("1" * (i + 1))
                # print(ctrl_state)
                mcp(beta, qv2, ctrl_state=ctrl_state)
            """ mcp(beta, qv2, ctrl_state = "0001")
            mcp(beta, qv2, ctrl_state = "0011")
            mcp(beta, qv2, ctrl_state = "0111")
            mcp(beta, qv2, ctrl_state = "1111") """

    return apply_mixer



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numba as nb
import numpy as np
import networkx as nx


def multi_hamiltonian_measurement(
    hamiltonians,
    qarg,
    precision=0.01,
    backend=None,
    shots=1000000,
    compile=True,
    compilation_kwargs={},
    subs_dic={},
    precompiled_qc=None,
    _measurements=None,
):
    r"""
    This method returns the expected value of a list of Hamiltonians for the state of a quantum argument.

    Parameters
    ----------
    hamiltonians : list[Hamiltonian]
        The Hamiltonians for wich the expeced value is to be evaluated.
    qarg : QuantumVariable, QuantumArray or list[QuantumVariable]
        The quantum argument to evaluate the Hamiltonians on.
    precision: float, optional
        The precision with which the expectation of the Hamiltonians is to be evaluated.
        The default is 0.01.
    backend : BackendClient, optional
        The backend on which to evaluate the quantum circuit. The default can be
        specified in the file default_backend.py.
    shots : integer, optional
        The maximum amount of shots to evaluate the expectation per Hamiltonian.
        The default is 100000.
    compile : bool, optional
        Boolean indicating if the .compile method of the underlying QuantumSession
        should be called before. The default is True.
    compilation_kwargs  : dict, optional
        Keyword arguments for the compile method. For more details check
        :meth:`QuantumSession.compile <qrisp.QuantumSession.compile>`. The default
        is ``{}``.
    subs_dic : dict, optional
        A dictionary of Sympy symbols and floats to specify parameters in the case
        of a circuit with unspecified, :ref:`abstract parameters<QuantumCircuit>`.
        The default is {}.
    precompiled_qc : QuantumCircuit, optional
            A precompiled quantum circuit.

    Returns
    -------
    list[float]
        The expected value of the Hamiltonians.

    """

    expectations = []
    n = len(hamiltonians)
    for i in range(n):
        expectations.append(
            hamiltonians[i].get_measurement(
                qarg,
                precision=precision,
                backend=backend,
                shots=shots,
                compile=compile,
                compilation_kwargs=compilation_kwargs,
                subs_dic=subs_dic,
                precompiled_qc=precompiled_qc,
                _measurement=None if _measurements == None else _measurements[i],
            )
        )

    return expectations


@nb.njit(cache=True)
def rlf_coloring(num_vertices, adjacency_matrix):
    colors = np.full(num_vertices, -1)
    uncolored_vertices = np.arange(num_vertices)

    current_color = 0
    while uncolored_vertices.size > 0:
        # Find the vertex with the maximum degree
        degrees = np.sum(adjacency_matrix[uncolored_vertices], axis=1)
        max_degree_index = np.argmax(degrees)
        independent_set = [uncolored_vertices[max_degree_index]]

        # Build the maximal independent set
        for vertex in uncolored_vertices:
            if np.all(adjacency_matrix[vertex, np.array(independent_set)] == 0):
                independent_set.append(vertex)

        # Assign the current color to all vertices in the independent set
        for vertex in independent_set:
            colors[vertex] = current_color

        # Remove colored vertices from the list of uncolored vertices
        uncolored_vertices = np.array(
            [v for v in uncolored_vertices if v not in independent_set]
        )

        current_color += 1

    return colors


@nb.njit(cache=True)
def dsatur_coloring(num_vertices, adjacency_matrix):
    colors = np.full(num_vertices, -1)
    saturation_degrees = np.zeros(num_vertices, dtype=np.int64)
    uncolored_vertices = np.arange(num_vertices)

    # Find the vertex with the maximum degree for the first coloring
    degrees = np.sum(adjacency_matrix, axis=1)
    max_degree_vertex = np.argmax(degrees)

    # Color the first vertex
    colors[max_degree_vertex] = 0
    uncolored_vertices = uncolored_vertices[uncolored_vertices != max_degree_vertex]

    # Update saturation degrees of neighbors
    for neighbor in range(num_vertices):
        if adjacency_matrix[max_degree_vertex, neighbor] == 1:
            saturation_degrees[neighbor] += 1

    while uncolored_vertices.size > 0:
        # Find the vertex with the highest saturation degree
        max_saturation = -1
        max_saturation_vertex = -1
        for vertex in uncolored_vertices:
            if saturation_degrees[vertex] > max_saturation:
                max_saturation = saturation_degrees[vertex]
                max_saturation_vertex = vertex
            elif saturation_degrees[vertex] == max_saturation:
                # If tied, choose the vertex with higher degree
                if degrees[vertex] > degrees[max_saturation_vertex]:
                    max_saturation_vertex = vertex

        # Find the lowest available color for this vertex
        used_colors = set()
        for neighbor in range(num_vertices):
            if (
                adjacency_matrix[max_saturation_vertex, neighbor] == 1
                and colors[neighbor] != -1
            ):
                used_colors.add(colors[neighbor])

        available_color = 0
        while available_color in used_colors:
            available_color += 1

        # Assign the color
        colors[max_saturation_vertex] = available_color

        # Update saturation degrees of uncolored neighbors
        for neighbor in uncolored_vertices:
            if adjacency_matrix[max_saturation_vertex, neighbor] == 1:
                neighbor_colors = set()
                for v in range(num_vertices):
                    if adjacency_matrix[neighbor, v] == 1 and colors[v] != -1:
                        neighbor_colors.add(colors[v])
                saturation_degrees[neighbor] = len(neighbor_colors)

        # Remove the colored vertex from uncolored_vertices
        uncolored_vertices = uncolored_vertices[
            uncolored_vertices != max_saturation_vertex
        ]

    return colors


def find_coloring(G):
    if len(G) == 0:
        return []

    adjacency_matrix = nx.to_numpy_array(G)

    coloring_1 = rlf_coloring(len(G), adjacency_matrix)
    coloring_2 = dsatur_coloring(len(G), adjacency_matrix)

    if np.max(coloring_1) < np.max(coloring_2):
        return coloring_1
    else:
        return coloring_2


def group_up_iterable(iterable, group_denominator):
    G = nx.Graph()

    for i in range(len(iterable)):
        G.add_node(i)
        for j in range(len(iterable)):
            if i == j:
                continue
            if not group_denominator(iterable[i], iterable[j]):
                G.add_edge(i, j)

    coloring = find_coloring(G)

    if len(coloring) == 0:
        return []

    groups = []
    for i in range(np.max(coloring) + 1):
        groups.append([])

    node_list = list(G.nodes())
    for i in range(len(G)):
        groups[coloring[i]].append(iterable[node_list[i]])

    return groups



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.core import QuantumVariable
from qrisp.qtypes import QuantumBool
from qrisp.environments import invert, custom_control
from qrisp.core.gate_application_functions import mcx, cx


# This function performs the Gidney adder from https://arxiv.org/pdf/1709.06648.pdf
@custom_control
def qq_gidney_adder(a, b, c_in=None, c_out=None, ctrl=None):

    if len(a) != len(b):
        raise Exception("Tried to call Gidney adder with inputs of unequal length")

    if c_out is not None:
        # Convert to qubit if neccessary
        if isinstance(c_out, QuantumBool):
            c_out = c_out[0]

        b = list(b) + [c_out]

    if len(b) == 1:
        if ctrl is not None:
            mcx([ctrl, a[0]], b)
        else:
            cx(a[0], b[0])
        if c_in is not None:
            if isinstance(c_in, QuantumBool):
                c_in = c_in[0]
            if ctrl is not None:
                cx(c_in, b[0])
            else:
                mcx([c_in, b[0]], b)
        return

    if ctrl is not None:
        gidney_control_anc = QuantumBool(name="gidney_control_anc*", qs=b[0].qs())

    gidney_anc = QuantumVariable(len(b) - 1, name="gidney_anc*", qs=b[0].qs())

    for i in range(len(b) - 1):

        if i != 0:
            mcx([a[i], b[i]], gidney_anc[i], method="gidney")
            cx(gidney_anc[i - 1], gidney_anc[i])
        elif c_in is not None:
            cx(c_in, b[i])
            cx(c_in, a[i])
            mcx([a[i], b[i]], gidney_anc[i], method="gidney")
            cx(c_in, gidney_anc[i])
        else:
            mcx([a[i], b[i]], gidney_anc[i], method="gidney")

        if i != len(b) - 2:
            cx(gidney_anc[i], a[i + 1])
            cx(gidney_anc[i], b[i + 1])

    if ctrl is not None:
        mcx([ctrl, gidney_anc[-1]], b[-1])

        mcx([ctrl, a[-1]], gidney_control_anc[0], method="gidney")
        cx(gidney_control_anc[0], b[-1])
        mcx([ctrl, a[-1]], gidney_control_anc[0], method="gidney_inv")

    else:
        cx(gidney_anc[-1], b[-1])

    with invert():
        for i in range(len(b) - 1):

            if i != 0:
                if i != len(b) - 1:
                    cx(gidney_anc[i - 1], a[i])

                if ctrl is not None:
                    if i != len(b) - 1:
                        cx(gidney_anc[i - 1], b[i])
                    mcx([ctrl, a[i]], gidney_control_anc[0], method="gidney")
                    cx(gidney_control_anc[0], b[i])
                    mcx([ctrl, a[i]], gidney_control_anc[0], method="gidney_inv")

                mcx([a[i], b[i]], gidney_anc[i], method="gidney")
                cx(gidney_anc[i - 1], gidney_anc[i])
            elif c_in is not None:
                cx(c_in, a[i])

                if ctrl is not None:
                    cx(c_in, b[i])
                    mcx([ctrl, c_in], gidney_control_anc[0], method="gidney")
                    cx(gidney_control_anc[0], b[i])
                    mcx([ctrl, c_in], gidney_control_anc[0], method="gidney_inv")

                mcx([a[i], b[i]], gidney_anc[i], method="gidney")
                cx(c_in, gidney_anc[i])
            else:

                if ctrl is not None:
                    mcx([ctrl, a[i]], gidney_control_anc[0], method="gidney")
                    cx(gidney_control_anc[0], b[i])
                    mcx([ctrl, a[i]], gidney_control_anc[0], method="gidney_inv")

                mcx([a[i], b[i]], gidney_anc[i], method="gidney")

    if ctrl is None:
        for i in range(len(a)):
            cx(a[i], b[i])
    else:
        gidney_control_anc.delete()

    gidney_anc.delete(verify=False)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

"""
This file implements the tools to perform quantum resource estimation using Jasp
infrastructure. The idea here is to transform the quantum instructions within a
given Jaspr into "counting instructions". That means instead of performing some
quantum gate, we increment an index in an array, which keeps track of how many
instructions of each type have been performed.

To do this, we implement the 

qrisp.jasp.interpreter_tools.interpreters.profiling_interpreter.py

Which handles the transformation logic of the Jaspr.
This file implements the interfaces to evaluating the transformed Jaspr.

"""


from functools import lru_cache

import numpy as np

from qrisp.jasp.interpreter_tools.abstract_interpreter import (
    extract_invalues,
    eval_jaxpr,
    insert_outvalues
)

from qrisp.jasp.primitives import (
    QuantumPrimitive,
    OperationPrimitive,
    AbstractQubitArray,
)


import jax
import jax.numpy as jnp
from jax.random import key



# This functions takes a "profiling dic", i.e. a dictionary of the form {str : int}
# indicating what kinds of quantum gates can appear in a Jaspr.
# It returns an equation evaluator, which increments a counter in an array for
# each quantum operation.
def make_profiling_eqn_evaluator(profiling_dic, meas_behavior):
    
    def profiling_eqn_evaluator(eqn, context_dic):

        invalues = extract_invalues(eqn, context_dic)

        if isinstance(eqn.primitive, QuantumPrimitive):

            # In the case of an OperationPrimitive, we determine the array index
            # to be increment via dictionary look-up and perform the increment
            # via the Jax-given .at method.
            if isinstance(eqn.primitive, OperationPrimitive):

                counting_array = list(invalues[-1][0])
                incrementation_constants = invalues[-1][1]

                op = eqn.primitive.op

                if op.definition:
                    op_counts = op.definition.transpile().count_ops()
                else:
                    op_counts = {op.name: 1}

                for op_name, count in op_counts.items():
                    counting_index = profiling_dic[op_name]
                    
                    # It seems like at this point we can just
                    # naively increment the Jax tracer but
                    # unfortunately the XLA compiler really 
                    # doesn't like constants. (Compile time blows up)
                    # We therefore jump through a lot of hoops
                    # to make it look like we are adding variables.
                    # This is done by supplying a list of tracers
                    # that will contain the numbers from 1 to n.
                    # Here is the next problem: If n is very large
                    # this also slows down the compilation because
                    # each function call has to have n+ arguments.
                    # We therefore perform the following loop:
                    while count:
                        incrementor = min(count, len(incrementation_constants))
                        count -= incrementor
                        counting_array[counting_index] += incrementation_constants[incrementor-1]

                insert_outvalues(eqn, context_dic, (counting_array, incrementation_constants))

            elif eqn.primitive.name == "jasp.measure":

                counting_index = profiling_dic["measure"]
                counting_array = list(invalues[-1][0])
                incrementation_constants = invalues[-1][1]

                meas_number = counting_array[counting_index]

                if isinstance(eqn.invars[0].aval, AbstractQubitArray):

                    def rng_body(i, acc):
                        meas_key = key(meas_number + i)
                        meas_res = meas_behavior(meas_key)

                        if (
                            not isinstance(meas_res, bool)
                            and not meas_res.dtype == jnp.bool
                        ):
                            raise Exception(
                                f"Tried to profil Jaspr with a measurement behavior not returning a boolean (got {meas_res.dtype}) instead"
                            )
                        acc = acc + (1 << i) * meas_res
                        return acc

                    meas_res = jax.lax.fori_loop(0, invalues[0], rng_body, jnp.int64(0))
                    counting_array[counting_index] += invalues[0]
                else:
                    meas_res = meas_behavior(key(meas_number))
                    if (
                        not isinstance(meas_res, bool)
                        and not meas_res.dtype == jnp.bool
                    ):
                        raise Exception(
                            f"Tried to profil Jaspr with a measurement behavior not returning a boolean (got {meas_res.dtype}) instead"
                        )
                        
                    counting_array[counting_index] += incrementation_constants[1]

                insert_outvalues(eqn, context_dic, [meas_res, (counting_array, incrementation_constants)])

            # Since we don't need to track to which qubits a certain operation
            # is applied, we can implement a really simple behavior for most
            # Qubit/QubitArray handling methods.
            # We represent qubit arrays simply with integers (indicating their)
            # size.
            elif eqn.primitive.name == "jasp.create_qubits":
                # create_qubits has the signature (size, QuantumCircuit).
                # Since we represent QubitArrays via integers, it is sufficient
                # to simply return the input as the output for this primitive.
                insert_outvalues(eqn, context_dic, invalues)

            elif eqn.primitive.name == "jasp.get_size":
                # The QubitArray size is represented via an integer.
                insert_outvalues(eqn, context_dic, invalues[0])

            elif eqn.primitive.name == "jasp.fuse":
                # The size of the fused qubit array is the size of the two added.
                insert_outvalues(eqn, context_dic, invalues[0] + invalues[1])

            elif eqn.primitive.name == "jasp.slice":
                # For the slice operation, we need to make sure, we don't go out
                # of bounds.
                start = jnp.max(jnp.array([invalues[1], 0]))
                stop = jnp.min(jnp.array([invalues[2], invalues[0]]))

                insert_outvalues(eqn, context_dic, stop - start)

            elif eqn.primitive.name == "jasp.get_qubit":
                # Trivial behavior since we don't need qubit address information
                insert_outvalues(eqn, context_dic, None)

            elif eqn.primitive.name in ["jasp.delete_qubits", "jasp.reset"]:
                # Trivial behavior: return the last argument (the counting array).
                insert_outvalues(eqn, context_dic, invalues[-1])
            elif eqn.primitive.name == "jasp.create_quantum_kernel":
                raise Exception("Tried to perform resource estimation on a function calling calling a kernelized function")
            else:
                raise Exception(
                    f"Don't know how to perform resource estimation with quantum primitive {eqn.primitive}"
                )

        elif eqn.primitive.name == "while":
            
            overall_constant_amount= max(eqn.params["body_nconsts"], eqn.params["cond_nconsts"])
            
            # Reinterpreted body and cond function
            def body_fun(val):
                
                constants = val[:eqn.params["body_nconsts"]]
                carries = val[overall_constant_amount:]
                
                body_res = eval_jaxpr(
                    eqn.params["body_jaxpr"], eqn_evaluator=profiling_eqn_evaluator
                )(*(constants + carries))
                
                return val[:overall_constant_amount] + tuple(body_res)

            def cond_fun(val):
                
                constants = val[:eqn.params["cond_nconsts"]]
                carries = val[overall_constant_amount:]
                
                res = eval_jaxpr(
                    eqn.params["cond_jaxpr"], eqn_evaluator=profiling_eqn_evaluator
                )(*(constants + carries))
                
                return res

            outvalues = jax.lax.while_loop(cond_fun, body_fun, tuple(invalues))[overall_constant_amount:]

            insert_outvalues(eqn, context_dic, outvalues)

        elif eqn.primitive.name == "cond":

            # Reinterpret branches
            branch_list = []

            for i in range(len(eqn.params["branches"])):
                branch_list.append(
                    eval_jaxpr(
                        eqn.params["branches"][i], eqn_evaluator=profiling_eqn_evaluator
                    )
                )

            outvalues = jax.lax.switch(invalues[0], branch_list, *invalues[1:])

            if len(eqn.outvars) == 1:
                outvalues = (outvalues,)

            insert_outvalues(eqn, context_dic, outvalues)

        elif eqn.primitive.name == "pjit":

            # For qached functions, we want to make sure, the compiled function
            # contains only a single implementation per qached function.

            # Within a Jaspr, it is made sure that qached function, which is called
            # multiple times only calls the Jaspr by reference in the pjit primitive
            # this way no "copies" of the implementation appear, thus keeping
            # the size of the intermediate representation limited.

            # We want to carry on this property. For this we use the lru_cache feature
            # on the get_compiler_profiler function. This function returns a
            # jitted function, which will always return the same object if called
            # with the same object. Since identical qached function calls are
            # represented by the same jaxpr, we achieve our goal.

            zipped_profiling_dic = tuple(profiling_dic.items())

            profiler = get_compiled_profiler(
                eqn.params["jaxpr"], zipped_profiling_dic, meas_behavior
            )

            outvalues = profiler(*invalues)

            if len(eqn.outvars) == 1:
                outvalues = (outvalues,)

            insert_outvalues(eqn, context_dic, outvalues)

        else:
            return True

    return profiling_eqn_evaluator


@lru_cache(int(1e5))
def get_compiled_profiler(jaxpr, zipped_profiling_dic, meas_behavior):

    profiling_dic = dict(zipped_profiling_dic)

    profiling_eqn_evaluator = make_profiling_eqn_evaluator(profiling_dic, meas_behavior)

    jitted_profiler = jax.jit(eval_jaxpr(jaxpr, eqn_evaluator=profiling_eqn_evaluator))
    
    call_counter = np.zeros(1)
    
    def profiler(*args):
        if call_counter[0] < 3 or len(jaxpr.eqns) < 20:
            call_counter[0] += 1
            return eval_jaxpr(jaxpr, eqn_evaluator=profiling_eqn_evaluator)(*args)
        else:
            return jitted_profiler(*args)

    return profiler



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.jasp.interpreter_tools import eval_jaxpr, extract_invalues, insert_outvalues


def pjit_to_gate(pjit_eqn, context_dic, eqn_evaluator):
    """
    Wraps the content of a pjit primitive into a gate and appends the gate.

    Parameters
    ----------
    pjit_eqn : An equation with a pjit primitive
        The equation to execute.

    Returns
    -------
    return values
        The values/tracers of the pjit execution.

    """

    # Set alias for the function definition
    definition_jaxpr = pjit_eqn.params["jaxpr"]

    # Extract the invalues from the context dic
    invalues = extract_invalues(pjit_eqn, context_dic)

    from qrisp.circuit import QuantumCircuit

    if len(invalues) == 0 or not isinstance(invalues[-1], QuantumCircuit):
        return True

    res = eval_qc(definition_jaxpr, invalues, eqn_evaluator)

    new_qc = res[-1]
    old_qc = invalues[-1]

    for i in range(len(new_qc.data)):
        if new_qc.data[i].op.num_clbits:
            old_qc.extend(new_qc, translation_dic = {b : b for b in new_qc.qubits + new_qc.clbits})
            break
    else:
        # Append the wrapped old circuit to the new circuit
        old_qc.append(
            new_qc.to_op(name=pjit_eqn.params["name"]), new_qc.qubits, new_qc.clbits
        )

    res = list(res)
    res[-1] = old_qc

    # Insert the result into the context dic
    insert_outvalues(pjit_eqn, context_dic, res)


def cond_to_cl_control(eqn, context_dic, eqn_evaluator):
    from qrisp.circuit import QuantumCircuit, Clbit

    if len(eqn.params["branches"]) > 2:
        raise Exception(
            "Tried to extract QuantumCircuit from Jaspr including a condition with more than two branches"
        )

    # Extract the invalues from the context dic
    invalues = extract_invalues(eqn, context_dic)

    if len(invalues) == 0:
        return True

    if not isinstance(invalues[0], Clbit):
        return True

    if not isinstance(invalues[-1], QuantumCircuit):
        raise Exception

    false_jaxpr = eqn.params["branches"][0]
    true_jaxpr = eqn.params["branches"][1]

    false_qc = eval_qc(false_jaxpr, invalues[1:], eqn_evaluator)[-1]
    true_qc = eval_qc(true_jaxpr, invalues[1:], eqn_evaluator)[-1]

    old_qc = invalues[-1]

    if len(false_qc.data):
        old_qc.append(
            false_qc.to_op("false_body").c_if(ctrl_state=0),
            false_qc.qubits,
            [invalues[0]] + false_qc.clbits,
        )
    if len(true_qc.data):
        old_qc.append(
            true_qc.to_op("true_body").c_if(ctrl_state=1),
            true_qc.qubits,
            [invalues[0]] + true_qc.clbits,
        )

    insert_outvalues(eqn, context_dic, [old_qc])


def eval_qc(definition_jaxpr, invalues, eqn_evaluator):

    old_qc = invalues[-1]
    new_qc = old_qc.clearcopy()
    invalues = list(invalues)
    invalues[-1] = new_qc

    res = eval_jaxpr(definition_jaxpr.jaxpr, eqn_evaluator=eqn_evaluator)(
        *(invalues + definition_jaxpr.consts)
    )

    if len(definition_jaxpr.jaxpr.outvars) == 1:
        res = [res]

    # Add new qubits/clbits to the circuit
    for qb in set(new_qc.qubits) - set(old_qc.qubits):
        old_qc.add_qubit(qb)

    for cb in set(new_qc.clbits) - set(old_qc.clbits):
        old_qc.add_clbit(cb)

    # Remove unused qu/clbits from the new circuit
    unused_qubits = set(new_qc.qubits)
    unused_clbits = set(new_qc.clbits)

    for instr in new_qc.data:
        unused_qubits -= set(instr.qubits)
        unused_clbits -= set(instr.clbits)

    for qb in unused_qubits:
        new_qc.qubits.remove(qb)
    for cb in unused_clbits:
        new_qc.clbits.remove(cb)

    res = list(res)
    res[-1] = new_qc

    return res



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import copy
from itertools import product

import numpy as np
import jax.numpy as jnp
import jax


from qrisp.circuit import transpile
from qrisp.core import QuantumVariable, qompiler, QuantumSession, merge
from qrisp.misc import bin_rep, lifted
from qrisp.jasp import (
    check_for_tracing_mode,
    q_fori_loop,
    jrange,
    create_qubits,
    DynamicQubitArray,
    TracingQuantumSession,
)


class QuantumArray:
    """
    This class allows the convenient management of multiple :ref:`QuantumVariables <QuantumVariable>` of one
    type. Inspired by the well known
    `numpy ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, the
    QuantumArray supports many convenient array manipulation methods. Similar to the
    numpy equivalent, creating a QuantumArray can be achieved by specifying a shape and
    a ``qtype``:

    >>> import numpy as np
    >>> from qrisp import QuantumArray, QuantumFloat
    >>> qtype = QuantumFloat(5, -2)
    >>> q_array = QuantumArray(qtype = qtype, shape = (2, 2, 2))

    Note that ``qtype`` is not a type object but a QuantumVariable which serves as an
    "example".

    To retrieve the entries (i.e. QuantumVariables) from the QuantumArray, we simply
    index as with regular numpy arrays:

    >>> from qrisp import h
    >>> qv = q_array[0,0,1]
    >>> h(qv[0])
    >>> print(q_array)
    {OutcomeArray([[[0., 0.],
                    [0., 0.]],
                   [[0., 0.],
                    [0., 0.]]]): 0.5,
     OutcomeArray([[[0.  , 0.25],
                    [0.  , 0.  ]],
                   [[0.  , 0.  ],
                    [0.  , 0.  ]]]): 0.5}

    We see the value 0.25 in the second entry because we applied an H-gate onto the 0-th
    qubit of the QuantumVariable at position (0,0,1). Since the type of this array is a
    QuantumFloat, with exponent -2, the significance of this qubit is 0.25.

    Note that the keys of the dictionary returned by the get_measurement method are no
    regular numpy arrays, as key objects need to be hashable. Instead, they are objects
    of an immutable subclass of np.ndarray called OutcomeArray, that supports hashing.

    For QuantumArrays, many methods known from numpy arrays work here too:

    >>> q_array = q_array.reshape(2,4)

    Not only do the ndarray methods work but also many other convenience functions from
    the numpy module:

    >>> q_array_swap = np.swapaxes(q_array, 0, 1)
    >>> print(q_array_swap)
    {OutcomeArray([[0., 0.],
                   [0., 0.],
                   [0., 0.],
                   [0., 0.]]): 0.5,
     OutcomeArray([[0.  , 0.  ],
                   [0.25, 0.  ],
                   [0.  , 0.  ],
                   [0.  , 0.  ]]): 0.5}

    To initiate the array, we use the :meth:`encode <qrisp.QuantumArray.encode>` method.
    Similar to QuantumVariables, we can also use the slicing operator, but this time
    non-trivial slices are possible as well:

    >>> q_array[1:,:] = 2*np.ones((1,4))
    >>> print(q_array)
    {OutcomeArray([[0., 0., 0., 0.],
                   [2., 2., 2., 2.]]): 0.5,
     OutcomeArray([[0.  , 0.25, 0.  , 0.  ],
                   [2.  , 2.  , 2.  , 2.  ]]): 0.5}


    **Quantum indexing**

    QuantumArrays can be dereferenced by :ref:`QuantumFloats <QuantumFloat>`. This
    returns a :ref:`QuantumEnvironment <QuantumEnvironment>` in which the corresponding entry is avaliable as
    a QuantumVariable. ::

        from qrisp import QuantumBool, QuantumArray, QuantumFloat, h, x, multi_measurement

        q_array = QuantumArray(QuantumFloat(1), shape = (4,4))
        index_0 = QuantumFloat(2)
        index_1 = QuantumFloat(2)


        index_0[:] = 2
        index_1[:] = 1

        h(index_0[0])

        with q_array[index_0, index_1] as entry:
            x(entry)

    >>> print(multi_measurement([index_0, index_1, q_array]))
    {(2, 1, OutcomeArray([[0., 0., 0., 0.],
                          [0., 0., 0., 0.],
                          [0., 1., 0., 0.],
                          [0., 0., 0., 0.]])): 0.5,
     (3, 1, OutcomeArray([[0., 0., 0., 0.],
                          [0., 0., 0., 0.],
                          [0., 0., 0., 0.],
                          [0., 1., 0., 0.]])): 0.5}

    .. note::
        This only works for arrays which have a size of an integer power of 2.

    **Matrix multiplication**

    For QuantumArrays with ``qtype`` QuantumFloat, matrix multiplication is available.

    >>> q_array_1 = QuantumArray(qtype, (2,2))
    >>> q_array_2 = QuantumArray(qtype, (2,2))
    >>> q_array_1[:] = 2*np.eye(2)
    >>> q_array_2[:] = [[1,2],[3,4]]
    >>> print(q_array_1 @ q_array_2)
    {OutcomeArray([[2., 4.],
                   [6., 0.]]): 1.0}

    .. note::
        By default, the output matrix will have the same ``qtype`` as the first input
        matrix. Here, the ``qtype`` is a QuantumFloat with 5 mantissa bits and exponent
        -2, implying that the result 8 yields overflow. Since qrisps unsigend arithmetic
        is modular, we get a 0.

    It is also possible to multiply classical and quantum matrices

    >>> q_array = QuantumArray(qtype, (2,2))
    >>> q_array[:] = 3*np.eye(2)
    >>> cl_array = np.array([[1,2],[3,4]])
    >>> print(q_array @ cl_array)
    {OutcomeArray([[3., 6.],
                   [1., 4.]]): 1.0}
    """

    def __init__(self, qtype, shape, qs=None):

        if isinstance(shape, int):
            shape = (shape,)
        size = 1
        for s in shape:
            if not isinstance(s, int):
                raise Exception(
                    f"Tried to create QuantumArray with non-integer tuple {shape}"
                )
            size *= s

        # The idea to implement this class with compatibility to dynamic features
        # (such as dynamic index support) is rooted in two core attributes:

        # 1. An integer jax array containing adresses
        # 2. A dynamic qubit array containig ALL qubits of the array

        # Many of the important properties (such as shape, size etc.) are derived
        # from the index array. Therefore manipulating these things can be achieved
        # by manipulating the index array.

        # If the user requests to retrieve a QuantumVariable from the QuantumArray,
        # the (dynamic) position of the corresponding qubits is retrieved from
        # the index array and from this a QuantumVariable is built up.
        # More on that in the __getitem__ method.

        self.ind_array = jnp.arange(size)
        self.ind_array = self.ind_array.reshape(shape)
        self.qtype = qtype

        if check_for_tracing_mode():

            if isinstance(qtype.reg, list):
                raise Exception(
                    "Tried to create QuantumArray with qtype defined outside of tracing context"
                )

            qs = qtype.qs
            self.qs = qs
            qb_array_tracer, qs.abs_qc = create_qubits(size * qtype.size, qs.abs_qc)
            self.qb_array = DynamicQubitArray(qb_array_tracer)
            self.qtype_size = qtype.size
        else:
            if qs is None:
                self.qs = QuantumSession()
            else:
                self.qs = qs
            self.qv_list = []
            for i in range(size):
                self.qv_list.append(
                    qtype.duplicate(name=self.qtype.name + "*", qs=self.qs)
                )

    @property
    def shape(self):
        return self.ind_array.shape

    @property
    def size(self):
        return self.ind_array.size

    @property
    def ndim(self):
        return self.ind_array.ndim

    def __getitem__(self, key):

        from qrisp.environments import conjugate

        # These cases represent the quantum indexing features
        if isinstance(key, QuantumVariable):
            merge([self.qs, key.qs])
            return conjugate(manipulate_array)(self, key)

        if isinstance(key, tuple):
            if all(isinstance(index, QuantumVariable) for index in key):
                merge([self.qs, key[0].qs])
                return conjugate(manipulate_array)(self, key)

        # If the key is not a tuple, convert to make further processing easier
        if not isinstance(key, tuple):
            key = (key,)

        # Retrieve the index address
        # This can either be an integer or an array slice, depending on what
        # the type of key is
        sliced_ind_array = self.ind_array[key]

        if len(sliced_ind_array.shape):
            # If the sliced_ind_array has a non trivial shape,
            # the result will be a QuantumArray (instead of a QuantumVariable).
            # We construct the sliced QuantumArray by copying all attributes
            # but instead use the sliced array as the index array.
            res = copy.copy(self)
            res.ind_array = sliced_ind_array
            return res
        else:
            # Otherwise the sliced_ind_array represents an integer indicating
            # the address.
            index = sliced_ind_array
            if check_for_tracing_mode():
                # To construct the resulting QuantumVariable we copy the qtype
                # variable and update the qv.reg attribute.
                qv = copy.copy(self.qtype)
                s = self.qtype_size
                qv.reg = self.qb_array[index * s : (index + 1) * s]
                qv.qs = self.qs
                return qv
            else:
                for i in range(len(key)):
                    if key[i] >= self.shape[i]:
                        raise Exception(
                            f"Index {key} out of bounds for QuantumArray with shape {self.shape}"
                        )
                return self.qv_list[index]

    def __setitem__(self, key, value):
        if isinstance(value, QuantumVariable):
            return
        sliced_array = self[key]
        sliced_array.encode(value)

    def encode(self, value):
        """
        The encode method allows to quickly bring a QuantumArray in a desired
        computational basis state. For this, it performs a circuit, bringing fresh
        qubits into the integer state specified by the encoder.

        A shorthand for this method is given by the ``[:]`` operator.

        Note that the qubits to initialize have to be fresh (i.e. no operations
        performed on them).

        Parameters
        ----------
        value :
            A value supported by the encoder.

        Examples
        --------

        We create a QuantumArray and encode the identity matrix.

        >>> from qrisp import QuantumArray, QuantumFloat
        >>> qtype = QuantumFloat(5)
        >>> q_array = QuantumArray(qtype, (4,4))
        >>> q_array.encode(np.eye(4))
        >>> print(q_array)
        {OutcomeArray([[1, 0, 0, 0],
                       [0, 1, 0, 0],
                       [0, 0, 1, 0],
                       [0, 0, 0, 1]]): 1.0}

        Using the slice operator we can also encode on slices of QuantumArrays

        >>> q_array = QuantumArray(qtype, (4,4))
        >>> q_array[:,:2] = np.ones((4,2))
        >>> print(q_array)
        {OutcomeArray([[1, 1, 0, 0],
                       [1, 1, 0, 0],
                       [1, 1, 0, 0],
                       [1, 1, 0, 0]]): 1.0}
        """

        if isinstance(value, list):
            value = np.array(value, dtype="object")

        if not value.shape == self.shape:
            raise Exception(
                "Tried to initialize a QuantumArray with incompatible shape"
            )

        flattened_value_array = value.flatten()
        flat_self = self.flatten()

        if check_for_tracing_mode() and isinstance(
            flattened_value_array, (np.ndarray, list)
        ):
            flattened_value_array = jnp.array(flattened_value_array)

        for i in jrange(self.size):
            flat_self[i][:] = flattened_value_array[i]

    def reshape(self, *args):
        """
        Adjusts the shape of the QuantumArray with similar semantics as
        `numpy.ndarray.reshape <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.reshape.html>`_.

        .. note::

            This method never allocates additional qubits and instead returns a
            `"view" <https://numpy.org/doc/2.2/user/basics.copies.html>`_,

        Parameters
        ----------
        shape : tuple
            The target shape.

        Returns
        -------
        res : QuantumArray
            The reshaped QuantumArray.

        Examples
        --------

        We create a 1-dimensional QuantumArray with $2**n$ entries and reshape it into
        a n dimensional QuantumArray with 2 entries per dimension.

        ::

            from qrisp import QuantumArray, QuantumFloat
            import numpy as np

            n = 3
            qtype = QuantumFloat(n)
            qa = QuantumArray(qtype = qtype, shape = 2**n)
            qa[:] = np.arange(2**n)

            print(qa)
            # Yields:
            # {OutcomeArray([0, 1, 2, 3, 4, 5, 6, 7]): 1.0}


        We can now reshape:

        ::

            reshaped_qa = qa.reshape(tuple(n*[2]))
            print(reshaped_qa)

            # Yields:
            # {OutcomeArray([[[0, 1],
            #                 [2, 3]],
            #                [[4, 5],
            #                 [6, 7]]]): 1.0}

        """
        if isinstance(args[0], (tuple, list)):
            shape = args[0]
        else:
            shape = args
        res = copy.copy(self)
        res.ind_array = self.ind_array.reshape(shape)
        return res

    def flatten(self):
        """
        Flattens the QuantumArray with similar semantics as
        `numpy.ndarray.flatten <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.flatten.html>`_.

        .. note::

            This method never allocates additional qubits and instead returns a
            `"view" <https://numpy.org/doc/2.2/user/basics.copies.html>`_,

        Returns
        -------
        res : QuantumArray
            The flattened QuantumArray.

        """

        return self.reshape(self.size)

    def ravel(self):
        """
        Ravels the QuantumArray with similar semantics as
        `numpy.ndarray.ravel <https://numpy.org/doc/stable/reference/generated/numpy.ravel.html>`_.

        .. note::

            This method never allocates additional qubits and instead returns a
            `"view" <https://numpy.org/doc/2.2/user/basics.copies.html>`_,

        Returns
        -------
        res : QuantumArray
            The raveled QuantumArray.

        """

        return self.flatten()

    def transpose(self, *axes):
        """
        Reverses the axes of the QuantumArray with similar semantics as
        `numpy.ndarray.transpose <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.transpose.html>`_.

        .. note::

            This method never allocates additional qubits and instead returns a
            `"view" <https://numpy.org/doc/2.2/user/basics.copies.html>`_.

        Parameters
        ----------
        *axes : None, tuple of ints, or n ints
                * None or no argument: reverses the order of the axes.
                * tuple of ints: i in the j-th place in the tuple means that the array’s i-th axis becomes the transposed array’s j-th axis.
                * n ints: same as an n-tuple of the same ints (this form is intended simply as a “convenience” alternative to the tuple form).

        Returns
        -------
        res : QuantumArray
            The transposed QuantumArray.

        """
        res = copy.copy(self)
        res.ind_array = self.ind_array.transpose(*axes)
        return res

    def swapaxes(self, axis_1, axis_2):
        """
        Swaps the axes of the QuantumArray with similar semantics as
        `numpy.ndarray.swapaxes <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.swapaxes.html>`_.

        .. note::

            This method never allocates additional qubits and instead returns a
            `"view" <https://numpy.org/doc/2.2/user/basics.copies.html>`_:

        Parameters
        ----------
        axis_1 : int
            First axis.
        axis_2 : int
            Second axis.

        Returns
        -------
        res : QuantumArray
            The QuantumArray with swapped axes.

        """
        res = copy.copy(self)
        res.ind_array = self.ind_array.swapaxes(axis_1, axis_2)
        return res

    def delete(self, verify=False):
        r"""
        Performs the :meth:`delete <qrisp.QuantumVariable.delete>` method on all
        QuantumVariables in this array.

        Parameters
        ----------
        verify : bool, optional
            If this keyword is set to true, a simulator is queried to check if the
            deleted qubits are in the $\ket{0}$ state. The default is False.

        """
        if check_for_tracing_mode():
            self.qs.clear_qubits(self.qb_array)
        else:
            for i in range(self.size):
                self.qv_list[i].delete(verify=verify)

    def measure(self):
        from qrisp import measure

        dtype = self.qtype.jdecoder(jnp.zeros(1)[0]).dtype

        meas_res = jnp.zeros(self.size, dtype=dtype)

        flattened_qa = self.flatten()

        def body_fun(i, val):
            meas_res, flattened_qa = val
            meas_res = meas_res.at[i].set(measure(flattened_qa[i]))
            return (meas_res, flattened_qa)

        meas_res, flattened_qa = q_fori_loop(
            0, flattened_qa.size, body_fun, (meas_res, flattened_qa)
        )

        return meas_res.reshape(self.shape)

    # Retrieves a measurement of the arrays state
    # Returns a list of tuples of the type (array, count)
    # ie. [(array([1,1,0]), 232), (array([1,1,3]), 115), ...]
    def get_measurement(
        self,
        backend=None,
        shots=None,
        compile=True,
        compilation_kwargs={},
        subs_dic={},
        circuit_preprocessor=None,
        precompiled_qc=None,
    ):
        """
        Method for acquiring measurement results for the given array. The semantics are
        similar to the :meth:`get_measurement <qrisp.QuantumVariable.get_measurement>`
        method of QuantumVariable. The results are returned as a dictionary of another
        numpy subtype called OutcomeArray.

        Parameters
        ----------
        backend : BackendClient, optional
            The backend on which to evaluate the quantum circuit. The default can be
            specified in the file default_backend.py.
        shots : integer, optional
            The amount of shots to evaluate the circuit. The default is given by the backend used.
        compile : bool, optional
            Boolean indicating if the .compile method of the underlying QuantumSession
            should be called before. The default is True.
        compilation_kwargs  : dict, optional
            Keyword arguments for the compile method. For more details check
            :meth:`QuantumSession.compile <qrisp.QuantumSession.compile>`.
            The default is ``{}``.
        subs_dic : dict, optional
            A dictionary of sympy symbols and floats to specify parameters in the case
            of a circuit with unspecified, abstract parameters. The default is {}.
        circuit_preprocessor : Python function, optional
            A function which recieves a QuantumCircuit and returns one, which is applied
            after compilation and parameter substitution. The default is None.

        Raises
        ------
        Exception
            Tried to get measurement within open environment.

        Returns
        -------
        list of tuples
            The measurement results in the form [(outcome_label, probability), ...].

        Examples
        --------

        >>> from qrisp import QuantumFloat, QuantumArray
        >>> qtype = QuantumFloat(3)
        >>> q_array = QuantumArray(qtype)
        >>> q_array[:] = [[1,0],[0,1]]
        >>> res = q_array.get_measurement()
        >>> print(res)
        {OutcomeArray([[1, 0],
                       [0, 1]]): 1.0}
        """

        if check_for_tracing_mode():
            raise Exception(
                "Tried to get_measurement from QuantumArray in tracing mode"
            )

        for qv in self.flatten():
            if qv.is_deleted():
                raise Exception(
                    "Tried to measure QuantumArray containing deleted QuantumVariables"
                )

        if backend is None:
            if self.qs.backend is None:
                from qrisp.default_backend import def_backend

                backend = def_backend
            else:
                backend = self.qs.backend

        if len(self.qs.env_stack) != 0:
            raise Exception("Tried to get measurement within open environment")

        qubits = sum([qv.reg for qv in self.flatten()[::-1]], [])
        # Copy circuit in over to prevent modification
        # from qrisp.quantum_network import QuantumNetworkClient

        if precompiled_qc is None:
            if compile:
                qc = qompiler(
                    self.qs, intended_measurements=qubits, **compilation_kwargs
                )
            else:
                qc = self.qs.copy()

            # Transpile circuit
            qc = transpile(qc)
        else:
            qc = precompiled_qc.copy()

        # Bind parameters
        if subs_dic:
            qc = qc.bind_parameters(subs_dic)
            from qrisp.core.compilation import combine_single_qubit_gates

            qc = combine_single_qubit_gates(qc)

        # Execute user specified circuit_preprocessor
        if circuit_preprocessor is not None:
            qc = circuit_preprocessor(qc)

        from qrisp.misc import get_measurement_from_qc

        counts = get_measurement_from_qc(qc, qubits, backend, shots)

        # Insert outcome labels (if available and hashable)
        new_counts_dic = {}
        for key in counts.keys():
            outcome_label = self.decoder(key)

            new_counts_dic[outcome_label] = counts[key]

        counts = new_counts_dic

        # Sort keys
        sorted_key_list = list(counts.keys())
        sorted_key_list.sort(key=lambda x: -counts[x])
        counts = {key: counts[key] for key in sorted_key_list}

        return counts

    def decoder(self, code_int):
        """
        The decoder method specifies how a QuantumArray turns the outcomes of
        measurements into human-readable values. It recieves an integer i and returns an
        OutcomeArray.

        Parameters
        ----------
        i : int
            Integer representing the outcome of a measurement of the qubits of this
            QuantumArray.

        Returns
        -------
        res : np.ndarray
            An array with entries of the type of the results of the .decoder of the
            qtype of this array.

        Examples
        --------

        We create a QuantumFloat and inspect its decoder:

        >>> from qrisp import QuantumArray, QuantumFloat
        >>> qtype = QuantumFloat(3)
        >>> q_array = QuantumArray(qtype, (2,2))
        >>> print(q_array.decoder(1))
        [[0 0]
         [0 1]]

        """

        flattened_array = self.flatten()

        from qrisp.qtypes.quantum_float import QuantumFloat

        if isinstance(self.qtype, QuantumFloat):
            if self.qtype.exponent >= 0:
                res = np.zeros(len(flattened_array), dtype=np.int32)
            else:
                res = np.zeros(len(flattened_array))
        else:
            res = np.zeros(len(flattened_array))

        n = len(self.qtype)

        bin_string = bin_rep(code_int, len(flattened_array) * n)

        for i in range(len(flattened_array)):
            if isinstance(self.qtype, QuantumFloat):
                res[i] = self.qtype.decoder(int(bin_string[i * n : (i + 1) * n], 2))
            else:
                res = res.astype("object")
                res[i] = self.qtype.decoder(int(bin_string[i * n : (i + 1) * n], 2))

        return OutcomeArray(res.reshape(self.shape))

    def __len__(self):
        return len(self.ind_array)

    def __str__(self):
        if not check_for_tracing_mode():
            return str(self.get_measurement())
        else:
            return "<QuantumArray[" + str(self.shape)[1:-1] + "]>"

    @lifted
    def __matmul__(self, other):
        from qrisp import QuantumFloat

        if isinstance(self.qtype, QuantumFloat):
            if isinstance(other, QuantumArray):
                from qrisp.alg_primitives.arithmetic import q_matmul

                return q_matmul(self, other)

            elif isinstance(other, np.ndarray):
                from qrisp.alg_primitives.arithmetic import semi_classic_matmul

                return semi_classic_matmul(self, other)

        raise Exception("Matrix multiplication for non-float types not implemented")

    def __rmatmul__(self, other):
        from qrisp import QuantumFloat

        if isinstance(self.qtype, QuantumFloat):
            return (self.transpose() @ other.transpose()).transpose()

    def __iter__(self):
        return QuantumArrayIterator(self.flatten())

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if ufunc is np.matmul:
            return inputs[1].__rmatmul__(inputs[0])
        return NotImplemented

    def concatenate(self, other, axis=0):
        """
        Concatenates two QuantumArrays along an axis with similar semantics as
        `numpy.concatenate <https://numpy.org/doc/stable/reference/generated/numpy.concatenate.html>`_.

        .. note::

            This method never allocates additional qubits and instead returns a
            `"view" <https://numpy.org/doc/2.2/user/basics.copies.html>`_.

        Parameters
        ----------
        other : QuantumArray
            The other array to concatenate.
        axis : int, optional
            The axis to concatenate along. The default is 0.

        Raises
        ------
        Exception
            Tried to concatenate two QuantumArrays with non-identical qtype.

        Returns
        -------
        res : QuantumArray
            The concatenated QuantumArray.

        """

        if not self.qtype is other.qtype:
            raise Exception(
                "Tried to concatenate two QuantumArrays with non-identical qtype"
            )

        res = copy.copy(self)

        ind_array_other_shifted = other.ind_array + self.size

        concat_ind_array = jnp.concatenate(
            (self.ind_array, ind_array_other_shifted), axis=axis
        )

        res.ind_array = concat_ind_array

        if check_for_tracing_mode():
            res.qb_array = self.qb_array + other.qb_array
        else:
            merge([self.qs, other.qs])
            res.qv_list = self.qv_list + other.qv_list

        return res

    def duplicate(self, init=False, qs=None):
        """
        This method returns a fresh QuantumArray, with equal ``qtype`` and shape.

        Parameters
        ----------
        init : bool, optional
            If set to True, the :meth:`init_from <qrisp.QuantumArray.init_from>` method
            will be called after creation. The default is False.
        qs : QuantumSession, optional
            The QuantumSession where the duplicate should be registered. By default,
            the duplicate will be registered in a new QuantumSession.

        Returns
        -------
        res : QuantumArray
            The duplicated array.

        Examples
        --------

        We duplicate a QuantumArray consisting of QuantumFloats with and without
        initiation.


        >>> from qrisp import QuantumArray, QuantumFloat
        >>> qtype = QuantumFloat(4)
        >>> q_array_0 = QuantumArray(qtype, (2,2))
        >>> q_array_0[:] = np.ones((2,2))
        >>> print(q_array_0)
        {OutcomeArray([[1, 1],
                       [1, 1]]): 1.0}
        >>> q_array_1 = q_array_0.duplicate()
        >>> print(q_array_1)
        {OutcomeArray([[0, 0],
                       [0, 0]]): 1.0}

        Note that no values have been carried over:

        >>> q_array_2 = q_array_0.duplicate(init = True)
        >>> print(q_array_2)
        {OutcomeArray([[1, 1],
                       [1, 1]]): 1.0}

        Now the values have been carried over. Note that this does NOT copy the state.
        For more information on this check the documentation of the
        :meth:`init_from <qrisp.QuantumVariable.init_from>` method of QuantumVariable.
        """

        res = copy.copy(self)

        if check_for_tracing_mode():
            qs = self.qs
            qb_array_tracer, qs.abs_qc = create_qubits(
                self.size * self.qtype_size, qs.abs_qc
            )
            res.qb_array = DynamicQubitArray(qb_array_tracer)

            if init:
                from qrisp import cx

                for i in jrange(self.size * self.qtype_size):
                    cx(self.qb_array[i], res.qb_array[i])

        else:

            if qs is None:
                res.qs = QuantumSession()
            else:
                res.qs = qs

            res.qv_list = []
            for i in range(self.size):
                res.qv_list.append(
                    self.qv_list[i].duplicate(
                        name=self.qtype.name + "*", qs=res.qs, init=init
                    )
                )

        return res

    def most_likely(self, **meas_kwargs):
        """
        Performs a measurement and returns the most likely outcome.

        Parameters
        ----------
        **kwargs : Keyword arguments for the get_measurement call.

        Examples
        --------

        >>> from qrisp import QuantumFloat, QuantumArray, ry
        >>> import numpy as np
        >>> qa = QuantumArray(QuantumFloat(3), shape = 4)
        >>> ry(np.pi*9/8, qa[0][0])
        >>> print(qa)
        {OutcomeArray([1, 0, 0, 0]): 0.9619, OutcomeArray([0, 0, 0, 0]): 0.0381}
        >>> qa.most_likely()
        OutcomeArray([1, 0, 0, 0])
        """
        meas_res = self.get_measurement(**meas_kwargs)
        return list(meas_res.keys())[0]


class QuantumArrayIterator:

    def __init__(self, qa):
        self.qa = qa
        self.counter = -1

    def __next__(self):
        self.counter += 1
        if self.counter >= self.qa.size:
            raise StopIteration
        return self.qa[self.counter]


def flatten_qa(qa):

    children = []

    qtype_children, qtype_aux_values = jax.tree.flatten(qa.qtype)

    children.append(qa.qtype_size)
    children.append(qa.ind_array)
    children.append(qa.qb_array)
    children.extend(qtype_children)

    aux_values = [qtype_aux_values]

    return tuple(children), tuple(aux_values)


def unflatten_qa(aux_data, children):

    qtype_children = children[3:]
    qtype_aux_values = aux_data[0]

    qtype = jax.tree.unflatten(qtype_aux_values, qtype_children)

    qa_dummy = object.__new__(QuantumArray)

    qtype_size = children[0]
    ind_array = children[1]
    qb_array = children[2]

    qa_dummy.qtype_size = qtype_size
    qa_dummy.qtype = qtype
    qa_dummy.ind_array = ind_array
    qa_dummy.qb_array = qb_array
    qa_dummy.qs = TracingQuantumSession.get_instance()

    return qa_dummy


jax.tree_util.register_pytree_node(QuantumArray, flatten_qa, unflatten_qa)


def manipulate_array(q_array, index):
    from qrisp import QuantumFloat, demux, invert

    if isinstance(index, tuple):
        if len(q_array.shape) != len(index):
            raise Exception(
                "Tried to quantum deref QuantumArray with index of mismatching shape"
            )

        for qf in index:
            if isinstance(qf, QuantumFloat):
                if qf.signed:
                    raise Exception("Tried to quantum deref with a signed QuantumFloat")
                if qf.exponent != 0:
                    raise Exception("Tried to quantum deref with a non-integer")

        index = sum([qv.reg for qv in index[::-1]], [])
        q_array = q_array.flatten()

    with invert():
        demux(q_array[0], index, q_array)

    return q_array[0]


class OutcomeArray(np.ndarray):
    def __new__(subtype, ndarray):
        if isinstance(ndarray, list):
            ndarray = np.array(ndarray)

            if ndarray.dtype == np.int64:
                ndarray = np.array(ndarray, dtype=np.int32)

        obj = super().__new__(subtype, ndarray.shape, dtype=ndarray.dtype)
        indices = product(*[list(range(i)) for i in ndarray.shape])
        for i in indices:
            np.ndarray.__setitem__(obj, i, ndarray[i])

        obj.flags.writeable = False
        return obj

    def __hash__(self):
        return hash(self.ravel().data.tobytes())
        # return hash(str(self))

    def __eq__(self, other):
        return np.array_equal(self, other)

    def __repr__(self):
        res = np.ndarray.__repr__(self).replace(", dtype=int32", "")

        return res



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp.jasp import qache, jrange
from qrisp.core import swap, h, cx, t, t_dg, s, p, measure, cz, cp, QuantumVariable
from qrisp.qtypes import QuantumBool, QuantumFloat
from qrisp.environments import control, custom_control, conjugate, invert
from qrisp.alg_primitives.arithmetic import gidney_adder


# @qache(static_argnames = "inpl_adder")
def jasp_mod_adder(a, b, modulus, inpl_adder=gidney_adder, ctrl=None):

    reduction_not_necessary = QuantumBool()
    # sign = QuantumBool()
    sign = b[-1]

    if isinstance(a, int):
        a = a % modulus

    # b = list(b) + [sign[0]]

    if ctrl is None:
        inpl_adder(a, b)
    else:
        with control(ctrl):
            inpl_adder(a, b)

    with invert():
        inpl_adder(modulus, b)

    cx(sign, reduction_not_necessary[0])

    with control(reduction_not_necessary[0]):
        inpl_adder(modulus, b)

    with invert():
        if ctrl is None:
            inpl_adder(a, b)
        else:
            with control(ctrl):
                inpl_adder(a, b)

    cx(sign, reduction_not_necessary[0])
    reduction_not_necessary.flip()

    if ctrl is None:
        inpl_adder(a, b)
    else:
        with control(ctrl):
            inpl_adder(a, b)

    reduction_not_necessary.delete()



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from functools import lru_cache

import numpy as np
from numba import njit
from jax.extend.core import ClosedJaxpr, JaxprEqn, Literal, Jaxpr


@lru_cache(maxsize=int(1e5))
def collect_environments(closed_jaxpr):
    """
    This function turns Jaxpr that contain QuantumEnvironment primitive in enter/exit
    form into the collected form. Collected means that the QuantumEnvironments content
    is represented by a Jaspr.

    Parameters
    ----------
    closed_jaxpr : jax.extend.core.ClosedJaxpr
        The Jaxpr with QuantumEnvironment in enter/exit form.

    Returns
    -------
    jax.extend.core.ClosedJaxpr
        A Jaxpr with QuantumEnvironments in collected form.

    """

    # We iterate through the list of equations, appending the equations to
    # the new list containing the processed equations.

    # Once we hit an exit primitive, we collect the Equations between the enter
    # and exit primitive.
    eqn_list = list(closed_jaxpr.jaxpr.eqns)
    new_eqn_list = []
    
    # An important part of collecting the quantum environments is determining
    # the input output variables. Doing this analysis can be prohibitvely costly
    # if implemented naively. For this reason the VarTracker class implements,
    # which tracks the I/O variables in a specialized data structured that
    # enables an efficient solution to this problem.
    eqn_var_tracker = VarTracker(eqn_list)
    new_eqn_var_tracker = VarTracker(new_eqn_list)
    
    from qrisp.jasp import Jaspr

    if isinstance(closed_jaxpr, Jaspr) and closed_jaxpr.envs_flattened:
        return closed_jaxpr
    
    for j in range(len(eqn_list)):
        eqn = eqn_list[j]

        if eqn.primitive.name == "pjit":

            new_params = dict(eqn.params)

            collected_jaspr = collect_environments(eqn.params["jaxpr"])

            new_params["jaxpr"] = collected_jaspr

            eqn = JaxprEqn(
                params=new_params,
                primitive=eqn.primitive,
                invars=list(eqn.invars),
                outvars=list(eqn.outvars),
                effects=eqn.effects,
                source_info=eqn.source_info,
                ctx=eqn.ctx
            )

        if eqn.primitive.name == "cond":

            new_params = dict(eqn.params)

            branch_list = []

            for i in range(len(eqn.params["branches"])):
                collected_branch_jaxpr = collect_environments(
                    eqn.params["branches"][i]
                )
                branch_list.append(collected_branch_jaxpr)

            new_params["branches"] = tuple(branch_list)

            eqn = JaxprEqn(
                params=new_params,
                primitive=eqn.primitive,
                invars=list(eqn.invars),
                outvars=list(eqn.outvars),
                effects=eqn.effects,
                source_info=eqn.source_info,
                ctx=eqn.ctx,
            )

        if eqn.primitive.name == "while":

            new_params = dict(eqn.params)

            body_collected_jaspr = collect_environments(eqn.params["body_jaxpr"])

            new_params["body_jaxpr"] = body_collected_jaspr

            eqn = JaxprEqn(
                params=new_params,
                primitive=eqn.primitive,
                invars=list(eqn.invars),
                outvars=list(eqn.outvars),
                effects=eqn.effects,
                source_info=eqn.source_info,
                ctx=eqn.ctx
            )

        # If an exit primitive is found, start the collecting mechanism.
        if eqn.primitive.name == "jasp.q_env" and "exit" in eqn.params.values():

            # Find the position of the enter primitive.
            for i in range(len(new_eqn_list))[::-1]:
                enter_eq = new_eqn_list[i]
                if (
                    enter_eq.primitive.name == "jasp.q_env"
                    and "enter" in enter_eq.params.values()
                ):
                    break
            else:
                raise

            # Set an alias for the equations marked as the body
            environment_body_eqn_list = new_eqn_list[i + 1 :]
            
            # Compute the sliced version of the var tracker
            environment_body_var_tracker = new_eqn_var_tracker.slice_start(i+1)
            
            # Compute the invars
            invars = environment_body_var_tracker.find_invars()
            
            # Remove the AbstractQuantumCircuit variable and prepend it.
            try:
                invars.remove(enter_eq.outvars[0])
            except ValueError:
                pass


            remaining_script_var_tracker = eqn_var_tracker.slice_start(j+1)
            
            # Same for the outvars
            outvars = find_outvars(
                environment_body_eqn_list,
                remaining_script_var_tracker,
                [var for var in closed_jaxpr.jaxpr.outvars if not isinstance(var, Literal)],
            )

            # Create the Jaxpr
            environment_body_jaspr = Jaspr(
                constvars=[],
                invars=invars + enter_eq.outvars,
                outvars=outvars + eqn.invars[-1:],
                eqns=environment_body_eqn_list,
            )

            # Create the Equation
            eqn = JaxprEqn(
                params={"type": eqn.params["type"], "jaspr": environment_body_jaspr},
                primitive=eqn.primitive,
                invars=enter_eq.invars[:-1] + invars + enter_eq.invars[-1:],
                outvars=outvars + eqn.outvars[-1:],
                effects=eqn.effects,
                source_info=eqn.source_info,
                ctx=eqn.ctx
            )

            # Remove the collected equations from the new_eqn_list
            new_eqn_list = new_eqn_list[:i]
            new_eqn_var_tracker = new_eqn_var_tracker.slice_end(i)
            

        # Append the equation
        new_eqn_list.append(eqn)
        new_eqn_var_tracker.append(eqn)
        
    if isinstance(closed_jaxpr, Jaspr):
        res = closed_jaxpr.update_eqns(new_eqn_list)
        
        if closed_jaxpr.ctrl_jaspr is not None:
            res.ctrl_jaspr = closed_jaxpr.ctrl_jaspr
        if closed_jaxpr.inv_jaspr is not None:
            res.inv_jaspr = closed_jaxpr.inv_jaspr
        return res
    else:
        # Return the transformed equation
        
        res_jaxpr = Jaxpr(
            constvars=closed_jaxpr.jaxpr.constvars,
            invars=closed_jaxpr.jaxpr.invars,
            outvars=closed_jaxpr.jaxpr.outvars,
            eqns=new_eqn_list,
        )
        
        return ClosedJaxpr(res_jaxpr, closed_jaxpr.consts)


def find_outvars(body_eqn_list, script_remainder_var_tracker, return_vars):
    """
    This function takes the equations of a function and some "follow-up"
    instructions and infers which variables need to be returned by the function.

    Parameters
    ----------
    body_eqn_list : list[JaxprEqn]
        A list of equations describing a function.
    script_remainder_eqn_list : list[JaxprEqn]
        A list of equations describing the follow up requirements.

    Returns
    -------
    list[jax.core.Var]
        A list of variables that would have to be returned by the function.

    """

    # This list will contain all variables produced by the function
    outvars = []

    # Fill the list
    for eqn in body_eqn_list:
        outvars.extend(eqn.outvars)

    # Remove the duplicates
    outvars = list(set(outvars))

    # Find which variables are required for executing the follow-up
    required_remainder_vars = script_remainder_var_tracker.find_invars()

    # The result is the intersection between both sets of variables
    return list(set(outvars).intersection(required_remainder_vars + return_vars))


class VarTracker:
    """
    This class is motivated by the task of identifying the inputs and outputs
    of the collected environments. For large Jaxpr, this can be prohibitively
    expensive, which is why this class tracks a specialized data structure, which
    allows an efficient solution of this problem.
    
    To describe how this works in detail, we start by noting that it is beneficial
    to assign every variable (input and outputs) a simply integer.
    
    The translation between this assignment is achieved through the var_to_int_dic
    and the int_to_var_dic.
    
    Given a list of equations, what this class tracks is now a list of integers 
    for both inputs and outputs that describe all the inputs and outputs.
    
    i.e. if we are given the equations
    
    %1 = prim_0 %2 %3 %4
    %5 %6 = prim_1 %1
    
    These lists would look like this:
        
    inputs = [%2, %3, %4, %1]
    outputs = [%1, %5, %6]
    
    We can now compute efficiently which variables are the invars of this list
    of equations by initializing an array with 6 entries, first setting all
    the invar positions to 1, i.e. [1, 1, 1, 1, 0, 0]
    and then setting all the outvar position to 0, i.e. [0, 1, 1, 1, 0, 0].

    This is implemented in the find_invars method.
    
    
    An important feature that is required by the environment collection function
    is to slice the list of equations - afterall the environment body will
    be a slice of the equation list in most cases.
    
    Because of this, the class tracks another list of integers, demarking
    which interval of integers belongs to which equation. This list always
    starts with 0.
    
    For the above example we would have
    
    invar_index_tracker = [0, 3, 4]
    outvar_index_tracker = [0, 1, 3]
    
    Each entry of these lists therefore denotes where the corresponding equation
    starts.
    
    This list can be used to efficiently implement the slicing features.
    """
    
    def __init__(self, eqn_list):
        
        
        # Initialize the translation dics
        var_to_int_dic = {}
        int_to_var_dic = {}
        
        # Initialize the variable lists
        eqn_invar_list = []
        eqn_outvar_list = []
        
        # Initialize the index trackers
        outvar_eqn_index_tracker = [0]
        invar_eqn_index_tracker = [0]
        
        # Fill the variable lists
        for eqn in eqn_list:
            invar_integers = []
            for var in eqn.invars:
                try:
                    invar_integers.append(var_to_int_dic[var])
                # This represents the case that the variable has not been
                # converted yet
                except KeyError:
                    var_to_int_dic[var] = len(var_to_int_dic)
                    int_to_var_dic[var_to_int_dic[var]] = var
                    invar_integers.append(var_to_int_dic[var])
                # This represents the case that the variable is a Literal
                except TypeError:
                    continue
            outvar_integers = []
            for var in eqn.outvars:
                try:
                    outvar_integers.append(var_to_int_dic[var])
                except KeyError:
                    var_to_int_dic[var] = len(var_to_int_dic)
                    int_to_var_dic[var_to_int_dic[var]] = var
                    outvar_integers.append(var_to_int_dic[var])
                except TypeError:
                    continue
            
            eqn_invar_list.extend(invar_integers)
            invar_eqn_index_tracker.append(len(eqn_invar_list))
            
            eqn_outvar_list.extend(outvar_integers)
            outvar_eqn_index_tracker.append(len(eqn_outvar_list))
            
        # Store the attributes
        self.var_to_int_dic = var_to_int_dic
        self.int_to_var_dic = int_to_var_dic
        
        self.eqn_invar_list = eqn_invar_list
        self.eqn_outvar_list = eqn_outvar_list
        
        self.invar_eqn_index_tracker = invar_eqn_index_tracker
        self.outvar_eqn_index_tracker = outvar_eqn_index_tracker
        

    def append(self, eqn):
        """
        Adds the equation eqn to the list of tracked equations.

        Parameters
        ----------
        eqn : jax.core.Equation

        """
        
        # Perform similar logic as in __init__        
        invar_integers = []
        for var in eqn.invars:
            try:
                invar_integers.append(self.var_to_int_dic[var])
            except KeyError:
                self.var_to_int_dic[var] = len(self.var_to_int_dic)
                self.int_to_var_dic[self.var_to_int_dic[var]] = var
                invar_integers.append(self.var_to_int_dic[var])
            except TypeError:
                continue
        outvar_integers = []
        for var in eqn.outvars:
            try:
                outvar_integers.append(self.var_to_int_dic[var])
            except KeyError:
                self.var_to_int_dic[var] = len(self.var_to_int_dic)
                self.int_to_var_dic[self.var_to_int_dic[var]] = var
                outvar_integers.append(self.var_to_int_dic[var])
            except TypeError:
                continue
    
        
        self.eqn_invar_list.extend(invar_integers)
        self.invar_eqn_index_tracker.append(len(self.eqn_invar_list))
        
        self.eqn_outvar_list.extend(outvar_integers)
        self.outvar_eqn_index_tracker.append(len(self.eqn_outvar_list))
        
        
    def slice_start(self, starting_point):
        """
        If self is represented by VarTracker(eqn_list), the result of this
        function is VarTracker(eqn_list[starting_point:])

        Parameters
        ----------
        starting_point : int
        """
        
        res = VarTracker([])
        
        # Identify where the invar list has to be sliced from
        invar_starting_point = self.invar_eqn_index_tracker[starting_point]
        # Slice the invar list
        res.eqn_invar_list = self.eqn_invar_list[invar_starting_point:]
        # Slice the index tracker and ensure it starts from 0
        res.invar_eqn_index_tracker = [i - invar_starting_point for i in self.invar_eqn_index_tracker[starting_point:]]
        
        # Same for the outvars
        outvar_starting_point = self.outvar_eqn_index_tracker[starting_point]
        res.eqn_outvar_list = self.eqn_outvar_list[outvar_starting_point:]
        res.outvar_eqn_index_tracker = [i - outvar_starting_point for i in self.outvar_eqn_index_tracker[starting_point:]]
        
        res.int_to_var_dic = self.int_to_var_dic
        res.var_to_int_dic = self.var_to_int_dic
        
        return res
    
    def slice_end(self, end_point):
        """
        If self is represented by VarTracker(eqn_list), the result of this
        function is VarTracker(eqn_list[:end_point])

        Parameters
        ----------
        end_point : int
        """
        
        res = VarTracker([])
        
        # Perform similar slicing logic as in slice_start
        invar_end_point = self.invar_eqn_index_tracker[end_point]
        res.eqn_invar_list = self.eqn_invar_list[:invar_end_point]
        res.invar_eqn_index_tracker = self.invar_eqn_index_tracker[:end_point+1]
        
        outvar_end_point = self.outvar_eqn_index_tracker[end_point]
        res.eqn_outvar_list = self.eqn_outvar_list[:outvar_end_point]
        res.outvar_eqn_index_tracker = self.outvar_eqn_index_tracker[:end_point+1]
        
        res.int_to_var_dic = self.int_to_var_dic
        res.var_to_int_dic = self.var_to_int_dic
        
        return res
    
    def find_invars(self):
        """
        Computes the undefined invars of the currently tracked equation list,
        i.e. all the variables that are used as invars but not defined by one
        of the equations.

        Returns
        -------
        res : list[Eqn]
        """
        
        # If viable, call the jitted version.
        if len(self.eqn_invar_list) < 20 or len(self.eqn_outvar_list) < 20:
            invar_index_list = find_invar_kernel([-1] + self.eqn_invar_list, [-1] +self.eqn_outvar_list)
        else:
            invar_index_list = jitted_find_invar_kernel(np.array([-1] + self.eqn_invar_list, dtype = np.int32), 
                                                        np.array([-1] +self.eqn_outvar_list, dtype = np.int32))
            
        res = []
        
        # Convert from integers to variables
        for i in range(len(invar_index_list)):
            res.append(self.int_to_var_dic[invar_index_list[i]])
        
        # For some jaxpr transformations, it is important that the order of invars
        # returned by this function is according to the order of appearance in the
        # body.
        
        # We therefore create a dictionary that indicates the index of the first
        # usage as an invar and sort according to this dictionary.
        sorting_dic = {self.eqn_invar_list[i] : i for i in range(len(self.eqn_invar_list))[::-1]}
        
        res.sort(key = lambda x : sorting_dic[self.var_to_int_dic[x]])
        
        return res
        
        
def find_invar_kernel(invar_indices, outvar_indices):
    # executes the algorithm described in the docstring of VarTracker
    
    max_invar = np.max(invar_indices)
    max_outvar = np.max(outvar_indices)
    max_var = max(max_invar, max_outvar)
    
    if max_var == -1:
        return np.zeros(0, dtype = np.int64)
    
    invar_array = np.zeros(max_var+1, dtype = np.int8)
    invar_array[invar_indices] = 1
    invar_array[outvar_indices] = 0
    res = np.nonzero(invar_array)[0]
    
    return res            

jitted_find_invar_kernel = njit(find_invar_kernel)        



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

"""
This file implements the tools to perform quantum resource estimation using Jasp
infrastructure. The idea here is to transform the quantum instructions within a
given Jaspr into "counting instructions". That means instead of performing some
quantum gate, we increment an index in an array, which keeps track of how many
instructions of each type have been performed.

To do this, we implement the 

qrisp.jasp.interpreter_tools.interpreters.profiling_interpreter.py

Which handles the transformation logic of the Jaspr.
This file implements the interfaces to evaluating the transformed Jaspr.

"""

from functools import lru_cache
import types

import numpy as np

import jax
import jax.numpy as jnp
from jax.extend.core import ClosedJaxpr


from qrisp.jasp.primitives import OperationPrimitive
from qrisp.jasp.interpreter_tools import make_profiling_eqn_evaluator, eval_jaxpr
from qrisp.jasp.evaluation_tools.jaspification import simulate_jaspr


def count_ops(meas_behavior):
    """
    Decorator to determine resources of large scale quantum computations.
    This decorator compiles the given Jasp-compatible function into a classical
    function computing the amount of each gates required. The decorated function
    will return a dictionary containing the operation counts.

    For many algorithms including classical feedback, the result of the
    measurements can heavily influence the required resources. To reflect this,
    users can specify the behavior of measurements during the computation of
    resources. The following strategies are available:

    * ``"0"`` - computes the resource as if measurements always return 0
    * ``"1"`` - computes the resource as if measurements always return 1
    * *callable* - allows the user to specify a random number generator (see examples)

    For more details on how the *callable* option can be used, consult the
    examples section.

    Finally it is also possible to call the Qrisp simulator to determine
    measurement behavior by providing ``sim``. This is of course much less
    scalable but in particular for algorithms involving repeat-until-success
    components, a necessary evil.

    Note that the ``sim`` option might return non-deterministic results, while
    the other methods do.

    .. warning::

        It is currently not possible to estimate programs, which include a
        :ref:`kernelized <quantum_kernel>` function.

    Parameters
    ----------
    meas_behavior : str or callable
        A string or callable indicating the behavior of the ressource computation
        when measurements are performed. Available are


    Returns
    -------
    resource_estimation decorator
        A decorator, producing a function to computed the required resources.

    Examples
    --------

    We compute the resources required to perform a large scale integer multiplication.

    ::

        from qrisp import count_ops, QuantumFloat, measure

        @count_ops(meas_behavior = "0")
        def main(i):

            a = QuantumFloat(i)
            b = QuantumFloat(i)

            c = a*b

            return measure(c)

        print(main(5))
        # {'cx': 506, 'x': 22, 'h': 135, 'measure': 55, '2cx': 2, 's': 45, 't': 90, 't_dg': 90}
        print(main(5000))
        # {'cx': 462552491, 'x': 20002, 'h': 112522500, 'measure': 37517500, '2cx': 2, 's': 37507500, 't': 75015000, 't_dg': 75015000}

    Note that even though the second computation contains more than 800 million gates,
    determining the resources takes less than 200ms, highlighting the scalability
    features of the Jasp infrastructure.

    **Modifying the measurement behavior via a random number generator**

    To specify the behavior, we specify an RNG function (for more details on
    what that means please check the `Jax documentation <https://docs.jax.dev/en/latest/jax.random.html>`_.
    This RNG takes as input a "key" and returns a boolean value.
    In this case, the return value will be uniformly distributed among True and False.

    ::


        from jax import random
        import jax.numpy as jnp
        from qrisp import QuantumFloat, measure, control, count_ops, x

        # Returns a uniformly distributed boolean
        def meas_behavior(key):
            return jnp.bool(random.randint(key, (1,), 0,1)[0])

        @count_ops(meas_behavior = meas_behavior)
        def main(i):

            qv = QuantumFloat(2)

            meas_res = measure(qv)

            with control(meas_res == i):
                x(qv)

            return measure(qv)

    This script executes two measurements and based on the measurement outcome
    executes two X gates. We can now execute this resource computation with
    different values of ``i`` to see, which measurements return ``True`` with
    our given random-number generator (recall that this way of specifying the
    measurement behavior is fully deterministic).

    ::

        print(main(0))
        # Yields: {'measure': 4, 'x': 2}
        print(main(1))
        # Yields: {'measure': 4}
        print(main(2))
        # Yields: {'measure': 4}
        print(main(3))
        # Yields: {'measure': 4}

    From this we conclude that our RNG returned 0 for both of the initial
    measurements.

    For some algorithms (such as :ref:`RUS`) sampling the measurement result
    from a simple distribution won't cut it because the required ressource can
    be heavily influenced by measurement outcomes. For this matter it is also
    possible to perform a full simulation. Note that this simulation is no
    longer deterministic.

    ::

        @count_ops(meas_behavior = "sim")
        def main(i):

            qv = QuantumFloat(2)

            meas_res = measure(qv)

            with control(meas_res == i):
                x(qv)

            return measure(qv)

        print(main(0))
        {'measure': 4, 'x': 2}
        print(main(1))
        {'measure': 4}

    """

    def count_ops_decorator(function):

        def ops_counter(*args):

            from qrisp.jasp import make_jaspr

            if not hasattr(function, "jaspr_dict"):
                function.jaspr_dict = {}

            args = list(args)

            signature = tuple([type(arg) for arg in args])
            if not signature in function.jaspr_dict:
                function.jaspr_dict[signature] = make_jaspr(function)(*args)

            return function.jaspr_dict[signature].count_ops(
                *args, meas_behavior=meas_behavior
            )

        return ops_counter

    return count_ops_decorator


def always_zero(key):
    return False


def always_one(key):
    return True


# This function is the central interface for performing resource estimation.
# It takes a Jaspr and returns a function, returning a dictionary (with the counted
# operations).
def profile_jaspr(jaspr, meas_behavior="0"):

    if isinstance(meas_behavior, str):
        if meas_behavior == "0":
            meas_behavior = always_zero
        elif meas_behavior == "1":
            meas_behavior = always_one
        elif not meas_behavior == "sim":
            raise Exception(
                f"Don't know how to compute required resources via method {meas_behavior}"
            )

    if callable(meas_behavior):

        def profiler(*args):

            # The profiling array computer is a function that computes the array
            # countaining the gate counts.
            # The profiling dic is a dictionary of type {str : int}, which indicates
            # which operation has been computed at which index of the array.
            profiling_array_computer, profiling_dic = get_profiling_array_computer(
                jaspr, meas_behavior
            )
            
            # Compute the profiling array
            if len(jaspr.outvars) > 1:
                profiling_array = profiling_array_computer(*args)[-1][0]
            else:
                profiling_array = profiling_array_computer(*args)[0]

            # Transform to a dictionary containing gate counts
            res_dic = {}
            for k in profiling_dic.keys():
                if int(profiling_array[profiling_dic[k]]):
                    res_dic[k] = int(profiling_array[profiling_dic[k]])

            return res_dic

    else:

        def profiler(*args):
            return simulate_jaspr(jaspr, *args, return_gate_counts=True)

    return profiler


# This function takes a Jaspr and returns a function computing the "counting array"
@lru_cache(int(1e5))
def get_profiling_array_computer(jaspr, meas_behavior):

    # This functions determines the set of primitives that appear in a given Jaxpr
    primitives = get_primitives(jaspr)

    # Filter out the non OperationPrimitives and fill them in a dictionary
    profiling_dic = {}
    for i in range(len(primitives)):
        if isinstance(primitives[i], OperationPrimitive):
            op = primitives[i].op
            if op.definition:
                op_names = list(op.definition.transpile().count_ops().keys())
            else:
                op_names = [op.name]
            for name in op_names:
                if not name in profiling_dic:
                    profiling_dic[name] = len(profiling_dic) - 1
        elif primitives[i].name == "jasp.measure" and not "measure" in profiling_dic:
            profiling_dic["measure"] = len(profiling_dic) - 1

    profiling_eqn_evaluator = make_profiling_eqn_evaluator(
        profiling_dic, meas_behavior
    )
    
    evaluator = jax.jit(eval_jaxpr(jaspr, eqn_evaluator=profiling_eqn_evaluator))

    # This function calls the profiling interpeter to evaluate the gate counts
    def profiling_array_computer(*args):

        # The XLA compiler showed some scalability problems in compile time.
        # Through a process involving a lot of blood and sweat
        # we reverse engineered what to do to improve these problems
        # 1. represent the integers that count the gates as a list
        # of integers (instead of an array)
        # 2. Avoid telling the compiler that it is constants that
        # are being added. To do this, we supply a list of the first
        # few integers as arguments, which will be used to do the
        # incrementation (i.e. CZ_count += 1). It therefore doesn't
        # look like a constant is being added but a variable
        final_arg = ([0] * len(profiling_dic), list(range(1, 6)))
        
        # Filter out types that are known to be static (https://github.com/eclipse-qrisp/Qrisp/issues/258)
        filtered_args = []        
        from qrisp.operators import QubitOperator, FermionicOperator
        
        for x in list(args) + [final_arg]:
            if type(x) not in [str, QubitOperator, FermionicOperator, types.FunctionType]:
                filtered_args.append(x)

        res = evaluator(*filtered_args)

        return res

    return profiling_array_computer, profiling_dic


# This functions determines the set of primitives that appear in a given Jaxpr
def get_primitives(jaxpr):

    primitives = set()

    for eqn in jaxpr.eqns:
        # Add current primitive
        primitives.add(eqn.primitive)

        if eqn.primitive.name == "cond":
            primitives.update(get_primitives(eqn.params["branches"][0].jaxpr))
            primitives.update(get_primitives(eqn.params["branches"][1].jaxpr))
            continue

        # Handle call primitives (like cond/pjit)
        for param in eqn.params.values():
            if isinstance(param, ClosedJaxpr):
                primitives.update(get_primitives(param.jaxpr))

    return list(primitives)



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import x, rx, rz, auto_uncompute, control


def qiro_rx_mixer(problem_updated):
    """
    RX-Mixer for QIRO algorithm. Works analogously to the normal RX-Mixer, but respects solutions and exclusions that have been found in the QIRO reduction steps.

    Parameters
    ----------
    solutions: List
        Solutions that have been found in the QIRO reduction steps.
    exclusions: List
        Solutions that have been found in the QIRO reduction steps.

    Returns
    -------
    RX_mixer: function
        The RX-mixer, according to the update steps that have been undertaken

    """

    solutions = problem_updated[1]
    exclusions = problem_updated[2]
    union = solutions + exclusions

    def RX_mixer(qv, beta):

        for i in range(len(qv)):
            if not i in union:
                rx(2 * beta, qv[i])

    return RX_mixer


def qiro_rz_mixer(problem_updated):
    """
    This function applies an RZ gate with a negative phase shift to a given quantum variable.

    Parameters
    ----------
    qv : QuantumVariable
        The quantum variable to which the RZ gate is applied.
    beta : float or sympy.Symbol
        The phase shift value for the RZ gate.

    """

    solutions = problem_updated[1]
    exclusions = problem_updated[2]
    union = solutions + exclusions

    def RZ_mixer(qv, beta):

        for i in range(len(qv)):
            if not i in union:
                rz(2 * beta, qv[i])

    return RZ_mixer


def qiro_controlled_RX_mixer_gen(predicate, union):
    r"""
    For a QIRO MIS instances, generate a controlled RX mixer for a given predicate function.

    Parameters
    ----------
    predicate : function
        A function receiving a ``QuantumVariable`` and an index $i$.
        This function returns a ``QuantumBool`` indicating if the predicate is satisfied for ``qv[i]``,
        that is, if the element ``qv[i]`` should be swapped in.
    union : List
        List of Qubits which were found to be positively or negatively correlated, i.e. they should not be mixed again.


    Returns
    -------
    controlled_RX_mixer : function
        A function receiving a ``QuantumVariable`` and a real parameter $\beta$.
        This function performs the application of the mixing operator.

    Examples
    --------

    We define the predicate function for the :ref:`MaxIndepSet <maxIndepSetQAOA>` problem. It returns ``True`` for the index (node) $i$ if
    all neighbors $j$ of the node $i$ in the graph $G$ are not selected, and ``False`` otherwise.

    ::

        from qrisp import QuantumVariable, QuantumBool, h, mcx, auto_uncompute, multi_measurement
        import networkx as nx

        G = nx.Graph()
        G.add_edges_from([(0, 1), (1, 2), (2, 0)])
        neighbors_dict = {node: list(G.adj[node]) for node in G.nodes()}

        def predicate(qv,i):
            qbl = QuantumBool()
            if len(neighbors_dict[i])==0:
                x(qbl)
            else:
                mcx([qv[j] for j in neighbors_dict[i]],qbl,ctrl_state='0'*len(neighbors_dict[i]))
            return qbl

        qv = QuantumVariable(3)
        h(qv)
        qbl = predicate(qv,0)
        multi_measurement([qv,qbl])
        # Yields: {('000', True): 0.125,('100', True): 0.125,('010', False): 0.125,('110', False): 0.125,('001', False): 0.125,('101', False): 0.125,('011', False): 0.125,('111', False): 0.125}

    The resulting ``controlled_RX_mixer`` then only swaps the node $i$ in if all neighbors $j$ in the graph $G$ are not selected.

    """

    @auto_uncompute
    def controlled_RX_mixer(qv, beta):
        m = qv.size
        for i in range(m):
            if i not in union:
                with control(predicate(qv, i)):
                    rx(beta, qv[i])

    return controlled_RX_mixer


def qiro_init_function(solutions=[], exclusions=[]):
    """
    State initialization function for QIRO algorithm. Works analogously to the normal initialization function, but respects solutions and exclusions that have been found in the QIRO reduction steps.

    Parameters
    ----------
    solutions: List
        Solutions that have been found in the QIRO reduction steps.
    exclusions: List
        Solutions that have been found in the QIRO reduction steps.

    Returns
    -------
    init_state: function
        The state initiation function, according to the update steps that have been undertaken

    """
    union = solutions + exclusions

    def init_state(qv):
        from qrisp import h

        # for i in problem.nodes:
        for i in range(len(qv)):
            if not i in union:
                h(qv[i])
        for i in solutions:
            x(qv[i])

    return init_state
